"""
Forecasting Model Base Classes

A forecasting model captures the future values of a time series as a function of the past as follows:

.. math:: y_{t+1} = f(y_t, y_{t-1}, ..., y_1),

where :math:`y_t` represents the time series' value(s) at time :math:`t`.

The main functions are `fit()` and `predict()`. `fit()` learns the function `f()`, over the history of
one or several time series. The function `predict()` applies `f()` on one or several time series in order
to obtain forecasts for a desired number of time stamps into the future.
"""
import copy
from typing import Optional, Tuple, Union, Any, Callable, Dict, List, Sequence
from itertools import product
from abc import ABC, ABCMeta, abstractmethod
from random import sample
import numpy as np
import pandas as pd

from darts.timeseries import TimeSeries
from darts.logging import get_logger, raise_log, raise_if_not, raise_if
from darts.utils import (
    _build_tqdm_iterator,
    _with_sanity_checks,
    _historical_forecasts_general_checks,
    _parallel_apply
)
from darts.utils.timeseries_generation import _generate_index
import inspect

from darts import metrics

logger = get_logger(__name__)


class ModelMeta(ABCMeta):
    def __call__(cls, *args, **kwargs):
        cls.model_call = (args, kwargs)
        return super(ModelMeta, cls).__call__(*args, **kwargs)


class ForecastingModel(ABC, metaclass=ModelMeta):
    """ The base class for forecasting models. It defines the *minimal* behavior that all forecasting models have to
        support. The signatures in this base class are for "local" models handling only one univariate series and no
        covariates. Sub-classes can handle more complex cases.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        # The series used for training the model through the `fit()` function.
        # This is only used if the model has been fit on one time series.
        self.training_series: Optional[TimeSeries] = None

        # state; whether the model has been fit (on a single time series)
        self._fit_called = False

        # extract and store sub class model creation parameters
        self._model_params = self._extract_model_creation_params()

    @abstractmethod
    def fit(self, series: TimeSeries) -> None:
        """ Trains the model on the provided univariate series.

        Parameters
        ----------
        series
            A target time series. The model will be trained to forecast this time series.
        """
        if not isinstance(self, DualCovariatesForecastingModel):
            series._assert_univariate()
        raise_if_not(len(series) >= self.min_train_series_length,
                     "Train series only contains {} elements but {} model requires at least {} entries"
                     .format(len(series), str(self), self.min_train_series_length))
        self.training_series = series
        self._fit_called = True

        if series.has_range_index:
            self._supports_range_index()

    def _supports_range_index(self) -> bool:
        """ Checks if the forecasting model supports a range index.
        Some models may not support this, if for instance the rely on underlying dates.

        By default, returns True. Needs to be overwritten by models that do not support
        range indexing and raise meaningful exception.
        """
        return True

    def _is_probabilistic(self) -> bool:
        """
        Checks if the forecasting model supports probabilistic predictions.
        By default, returns False. Needs to be overwritten by models that do support
        probabilistic predictions.
        """
        return False

    def _supports_non_retrainable_historical_forecasts(self) -> bool:
        """
        Checks if the forecasting model supports historical forecasts without retraining
        the model. By default, returns False. Needs to be overwritten by models that do
        support historical forecasts without retraining.
        """
        return False


    @property
    def uses_past_covariates(self):
        return 'past_covariates' in inspect.signature(self.fit).parameters.keys()

    @property
    def uses_future_covariates(self):
        return 'future_covariates' in inspect.signature(self.fit).parameters.keys()

    @abstractmethod
    def predict(self,
                n: int,
                num_samples: int = 1) -> TimeSeries:
        """ Forecasts values for `n` time steps after the end of the series.

        Parameters
        ----------
        n
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.

        Returns
        -------
        TimeSeries
            A time series containing the `n` next points after then end of the training series.
        """
        if not self._fit_called:
            raise_log(ValueError('The model must be fit before calling `predict()`.'
                                 'For global models, if `predict()` is called without specifying a series,'
                                 'the model must have been fit on a single training series.'), logger)

        if not self._is_probabilistic() and num_samples > 1:
            raise_log(ValueError('`num_samples > 1` is only supported for probabilistic models.'), logger)

    def _fit_wrapper(self, series: TimeSeries,
                     past_covariates: Optional[TimeSeries],
                     future_covariates: Optional[TimeSeries]):
        self.fit(series)

    def _predict_wrapper(self, n: int, series: TimeSeries,
                         past_covariates: Optional[TimeSeries],
                         future_covariates: Optional[TimeSeries],
                         num_samples: int) -> TimeSeries:
        return self.predict(n, num_samples=num_samples)

    @property
    def min_train_series_length(self) -> int:
        """
        Class property defining the minimum required length for the training series.
        This function/property should be overridden if a value higher than 3 is required.
        """
        return 3

    def _generate_new_dates(self,
                            n: int,
                            input_series: Optional[TimeSeries] = None) -> Union[pd.DatetimeIndex, pd.RangeIndex]:
        """
        Generates `n` new dates after the end of the specified series
        """
        input_series = input_series if input_series is not None else self.training_series

        last = input_series.end_time()
        start = last + input_series.freq if input_series.has_datetime_index else last + 1
        return _generate_index(start=start, freq=input_series.freq, length=n)

    def _build_forecast_series(self,
                               points_preds: Union[np.ndarray, Sequence[np.ndarray]],
                               input_series: Optional[TimeSeries] = None) -> TimeSeries:
        """
        Builds a forecast time series starting after the end of the training time series, with the
        correct time index (or after the end of the input series, if specified).
        """
        input_series = input_series if input_series is not None else self.training_series
        time_index_length = len(points_preds) if isinstance(points_preds, np.ndarray) else len(points_preds[0])
        time_index = self._generate_new_dates(time_index_length, input_series=input_series)
        if isinstance(points_preds, np.ndarray):
            return TimeSeries.from_times_and_values(time_index,
                                                    points_preds,
                                                    freq=input_series.freq_str,
                                                    columns=input_series.columns)

        return TimeSeries.from_times_and_values(time_index,
                                                np.stack(points_preds, axis=2),
                                                freq=input_series.freq_str,
                                                columns=input_series.columns)

    def _historical_forecasts_sanity_checks(self, *args: Any, **kwargs: Any) -> None:
        """Sanity checks for the historical_forecasts function

        Parameters
        ----------
        args
            The args parameter(s) provided to the historical_forecasts function.
        kwargs
            The kwargs paramter(s) provided to the historical_forecasts function.

        Raises
        ------
        ValueError
            when a check on the parameter does not pass.
        """
        # parse args and kwargs
        series = args[0]
        _historical_forecasts_general_checks(series, kwargs)

    def _get_last_prediction_time(self, series, forecast_horizon, overlap_end):
        if overlap_end:
            last_valid_pred_time = series.time_index[-1]
        else:
            last_valid_pred_time = series.time_index[-forecast_horizon]

        return last_valid_pred_time

    @_with_sanity_checks("_historical_forecasts_sanity_checks")
    def historical_forecasts(self,
                             series: TimeSeries,
                             past_covariates: Optional[TimeSeries] = None,
                             future_covariates: Optional[TimeSeries] = None,
                             num_samples: int = 1,
                             start: Union[pd.Timestamp, float, int] = 0.5,
                             forecast_horizon: int = 1,
                             stride: int = 1,
                             retrain: bool = True,
                             overlap_end: bool = False,
                             last_points_only: bool = True,
                             verbose: bool = False) -> Union[TimeSeries, List[TimeSeries]]:

        """
        Computes the historical forecasts the model would have produced with an expanding training window
        and (by default) returns a time series created from the last point of each of these individual forecasts.
        To this end, it repeatedly builds a training set from the beginning of `series`. It trains the
        current model on the training set, emits a forecast of length equal to forecast_horizon, and then moves
        the end of the training set forward by `stride` time steps.

        By default, this method will return a single time series made up of the last point of each
        historical forecast. This time series will thus have a frequency of `series.freq * stride`.
        If `last_points_only` is set to False, it will instead return a list of the historical forecasts.

        By default, this method always re-trains the models on the entire available history,
        corresponding to an expanding window strategy.
        If `retrain` is set to False (useful for models for which training might be time-consuming, such as
        deep learning models), the model will only be trained on the initial training window
        (up to `start` time stamp), and only if it has not been trained before. Then, at every iteration, the
        newly expanded input sequence will be fed to the model to produce the new output.

        Parameters
        ----------
        series
            The target time series to use to successively train and evaluate the historical forecasts.
        past_covariates
            An optional past-observed covariate series. This applies only if the model supports past covariates.
        future_covariates
            An optional future-known covariate series. This applies only if the model supports future covariates.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.
        start
            The first point of time at which a prediction is computed for a future time.
            This parameter supports 3 different data types: `float`, `int` and `pandas.Timestamp`.
            In the case of `float`, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of `int`, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            In case of `pandas.Timestamp`, this time stamp will be used to determine the first prediction time
            directly.
        forecast_horizon
            The forecast horizon for the predictions
        stride
            The number of time steps between two consecutive predictions.
        retrain
            Whether to retrain the model for every prediction or not. Currently only `TorchForecastingModel`
            instances such as `BlockRNNModel`, `RNNModel`, `TCNModel`, `NBEATSModel` and `TransformerModel` support
            setting `retrain` to `False`.
        overlap_end
            Whether the returned forecasts can go beyond the series' end or not
        last_points_only
            Whether to retain only the last point of each historical forecast.
            If set to True, the method returns a single `TimeSeries` containing the successive point forecasts.
            Otherwise returns a list of historical `TimeSeries` forecasts.
        verbose
            Whether to print progress
        Returns
        -------
        TimeSeries or List[TimeSeries]
            By default, a single TimeSeries instance created from the last point of each individual forecast.
            If `last_points_only` is set to False, a list of the historical forecasts.
        """

        # TODO: do we need a check here? I'd rather leave these checks to the models/datasets.
        # if covariates:
        #     raise_if_not(series.end_time() <= covariates.end_time() and covariates.start_time() <= series.start_time(),
        #                  'The provided covariates must be at least as long as the target series.')

        # only GlobalForecastingModels support historical forecastings without retraining the model
        base_class_name = self.__class__.__base__.__name__
        raise_if(not retrain and not self._supports_non_retrainable_historical_forecasts(),
                 f'{base_class_name} does not support historical forecastings with `retrain` set to `False`. '
                 f'For now, this is only supported with GlobalForecastingModels such as TorchForecastingModels. '
                 f'Fore more information, read the documentation for `retrain` in `historical_forecastings()`',
                 logger)

        # prepare the start parameter -> pd.Timestamp
        start = series.get_timestamp_at_point(start)

        # build the prediction times in advance (to be able to use tqdm)
        last_valid_pred_time = self._get_last_prediction_time(series, forecast_horizon, overlap_end)

        pred_times = [start]
        while pred_times[-1] < last_valid_pred_time:
            # compute the next prediction time and add it to pred times
            pred_times.append(pred_times[-1] + series.freq * stride)

        # the last prediction time computed might have overshot last_valid_pred_time
        if pred_times[-1] > last_valid_pred_time:
            pred_times.pop(-1)

        iterator = _build_tqdm_iterator(pred_times, verbose)

        # Either store the whole forecasts or only the last points of each forecast, depending on last_points_only
        forecasts = []

        last_points_times = []
        last_points_values = []

        # iterate and forecast
        for pred_time in iterator:
            train = series.drop_after(pred_time)  # build the training series

            # train_cov = covariates.drop_after(pred_time) if covariates else None

            if retrain:
                self._fit_wrapper(series=train, past_covariates=past_covariates, future_covariates=future_covariates)

            forecast = self._predict_wrapper(n=forecast_horizon,
                                             series=train,
                                             past_covariates=past_covariates,
                                             future_covariates=future_covariates,
                                             num_samples=num_samples)

            if last_points_only:
                last_points_values.append(forecast.all_values()[-1])
                last_points_times.append(forecast.end_time())
            else:
                forecasts.append(forecast)

        if last_points_only:
            if series.has_datetime_index:
                return TimeSeries.from_times_and_values(pd.DatetimeIndex(last_points_times, freq=series.freq * stride),
                                                        np.array(last_points_values))
            else:
                return TimeSeries.from_times_and_values(pd.RangeIndex(start=last_points_times[0],
                                                                      stop=last_points_times[-1] + 1,
                                                                      step=1),
                                                        np.array(last_points_values))


        return forecasts

    def backtest(self,
                 series: TimeSeries,
                 past_covariates: Optional[TimeSeries] = None,
                 future_covariates: Optional[TimeSeries] = None,
                 num_samples: int = 1,
                 start: Union[pd.Timestamp, float, int] = 0.5,
                 forecast_horizon: int = 1,
                 stride: int = 1,
                 retrain: bool = True,
                 overlap_end: bool = False,
                 last_points_only: bool = False,
                 metric: Callable[[TimeSeries, TimeSeries], float] = metrics.mape,
                 reduction: Union[Callable[[np.ndarray], float], None] = np.mean,
                 verbose: bool = False) -> Union[float, List[float]]:

        """
        Computes error scores between the historical forecasts the model would have produced
        with an expanding training window over `series` and the actual values.
        To this end, it repeatedly builds a training set from the beginning of `series`. It trains the current model on
        the training set, emits a forecast of length equal to forecast_horizon, and then moves the end of the
        training set forward by `stride` time steps. A metric (given by the `metric` function) is then evaluated
        on the forecast and the actual values. Finally, the method returns a `reduction` (the mean by default)
        of all these metric scores.

        By default, this method will use each historical forecast (whole) to compute error scores.
        If `last_points_only` is set to True, it will use only the last point of each historical
        forecast. In this case, no reduction is used.

        By default, this method always re-trains the models on the entire available history,
        corresponding to an expanding window strategy.
        If `retrain` is set to False (useful for models for which training might be time-consuming, such as
        deep learning models), the model will only be trained on the initial training window
        (up to `start` time stamp), and only if it has not been trained before. Then, at every iteration, the
        newly expanded input sequence will be fed to the model to produce the new output.

        Parameters
        ----------
        series
            The target time series to use to successively train and evaluate the historical forecasts
        past_covariates
            An optional past-observed covariate series. This applies only if the model supports past covariates.
        future_covariates
            An optional future-known covariate series. This applies only if the model supports future covariates.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.
        start
            The first prediction time, at which a prediction is computed for a future time.
            This parameter supports 3 different data types: `float`, `int` and `pandas.Timestamp`.
            In the case of `float`, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            In the case of `int`, the parameter will be treated as an integer index to the time index of
            `training_series` that will be used as first prediction time.
            In case of `pandas.Timestamp`, this time stamp will be used to determine the first prediction time
            directly.
        forecast_horizon
            The forecast horizon for the point prediction
        stride
            The number of time steps between two consecutive predictions.
        retrain
            Whether to retrain the model for every prediction or not. Currently only `TorchForecastingModel`
            instances such as `BlockRNNModel`, `RNNModel`, `TCNModel`, `NBEATSModel` and `TransformerModel` support
            setting `retrain` to `False`.
        overlap_end
            Whether the returned forecasts can go beyond the series' end or not
        last_points_only
            Whether to use the whole historical forecasts or only the last point of each forecast to compute the error
        metric
            A function that takes two TimeSeries instances as inputs and returns a float error value.
        reduction
            A function used to combine the individual error scores obtained when `last_points_only` is set to False.
            If explicitely set to `None`, the method will return a list of the individual error scores instead.
            Set to np.mean by default.
        verbose
            Whether to print progress
        Returns
        -------
        float or List[float]
            The error score, or the list of individual error scores if `reduction` is `None`
        """
        forecasts = self.historical_forecasts(series=series,
                                              past_covariates=past_covariates,
                                              future_covariates=future_covariates,
                                              num_samples=num_samples,
                                              start=start,
                                              forecast_horizon=forecast_horizon,
                                              stride=stride,
                                              retrain=retrain,
                                              overlap_end=overlap_end,
                                              last_points_only=last_points_only,
                                              verbose=verbose)

        if last_points_only:
            return metric(series, forecasts)

        errors = [metric(series, forecast) for forecast in forecasts]
        if reduction is None:
            return errors

        return reduction(np.array(errors))

    @classmethod
    def gridsearch(model_class,
                   parameters: dict,
                   series: TimeSeries,
                   past_covariates: Optional[TimeSeries] = None,
                   future_covariates: Optional[TimeSeries] = None,
                   forecast_horizon: Optional[int] = None,
                   stride: int = 1,
                   start: Union[pd.Timestamp, float, int] = 0.5,
                   last_points_only: bool = False,
                   val_series: Optional[TimeSeries] = None,
                   use_fitted_values: bool = False,
                   metric: Callable[[TimeSeries, TimeSeries], float] = metrics.mape,
                   reduction: Callable[[np.ndarray], float] = np.mean,
                   verbose=False,
                   n_jobs: int = 1,
                   n_random_samples: Optional[Union[int, float]] = None) -> Tuple['ForecastingModel', Dict]:
        """
        A function for finding the best hyper-parameters among a given set.
        This function has 3 modes of operation: Expanding window mode, split mode and fitted value mode.
        The three modes of operation evaluate every possible combination of hyper-parameter values
        provided in the `parameters` dictionary by instantiating the `model_class` subclass
        of ForecastingModel with each combination, and returning the best-performing model with regards
        to the `metric` function. The `metric` function is expected to return an error value,
        thus the model resulting in the smallest `metric` output will be chosen.

        The relationship of the training data and test data depends on the mode of operation.

        Expanding window mode (activated when `forecast_horizon` is passed):
        For every hyperparameter combination, the model is repeatedly trained and evaluated on different
        splits of `training_series` and `target_series`. This process is accomplished by using
        the `backtest` function as a subroutine to produce historic forecasts starting from `start`
        that are compared against the ground truth values of `training_series` or `target_series`, if
        specified. Note that the model is retrained for every single prediction, thus this mode is slower.

        Split window mode (activated when `val_series` is passed):
        This mode will be used when the `val_series` argument is passed.
        For every hyper-parameter combination, the model is trained on `series` and
        evaluated on `val_series`.

        Fitted value mode (activated when `use_fitted_values` is set to `True`):
        For every hyper-parameter combination, the model is trained on `series`
        and evaluated on the resulting fitted values.
        Not all models have fitted values, and this method raises an error if the model doesn't have a `fitted_values`
        member. The fitted values are the result of the fit of the model on `series`. Comparing with the
        fitted values can be a quick way to assess the model, but one cannot see if the model is overfitting the series.

        Derived classes must ensure that a single instance of a model will not share parameters with the other
        instances, e.g., saving models in the same path. Otherwise, an unexpected behavior can arise while running
        several models in parallel (when `n_jobs != 1`). If this cannot be avoided, then gridsearch should be redefined,
        forcing `n_jobs = 1`.

        Currently this method only supports deterministic predictions (i.e. `num_samples == 1`).

        Parameters
        ----------
        model_class
            The ForecastingModel subclass to be tuned for 'series'.
        parameters
            A dictionary containing as keys hyperparameter names, and as values lists of values for the
            respective hyperparameter.
        series
            The TimeSeries instance used as input and target for training.
        past_covariates
            An optional past-observed covariate series. This applies only if the model supports past covariates.
        future_covariates
            An optional future-known covariate series. This applies only if the model supports future covariates.
        forecast_horizon
            The integer value of the forecasting horizon used in expanding window mode.
        stride
            The number of time steps between two consecutive predictions. Only used in expanding window mode.
        start
            The `int`, `float` or `pandas.Timestamp` that represents the starting point in the time index
            of `training_series` from which predictions will be made to evaluate the model.
            For a detailed description of how the different data types are interpreted, please see the documentation
            for `ForecastingModel.backtest`.
        last_points_only
            Whether to use the whole forecasts or only the last point of each forecast to compute the error
        val_series
            The TimeSeries instance used for validation in split mode. If provided, this series must start right after
            the end of `series`; so that a proper comparison of the forecast can be made.
        use_fitted_values
            If `True`, uses the comparison with the fitted values.
            Raises an error if `fitted_values` is not an attribute of `model_class`.
        metric
            A function that takes two TimeSeries instances as inputs and returns a float error value.
        reduction
            A reduction function (mapping array to float) describing how to aggregate the errors obtained
            on the different validation series when backtesting. By default it'll compute the mean of errors.
        verbose
            Whether to print progress.
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when there are two or more parameters
            combinations to evaluate. Each job will instantiate, train, and evaluate a different instance of the model.
            Defaults to `1` (sequential). Setting the parameter to `-1` means using all the available cores.
        n_random_samples
            The number/ratio of hyperparameter combinations to select from the full parameter grid. This will perform
            a random search instead of using the full grid.
            If an integer, `n_random_samples` is the number of parameter combinations selected from the full grid and must
            be between `0` and the total number of parameter combinations.
            If a float, `n_random_samples` is the ratio of parameter combinations selected from the full grid and must be
            between `0` and `1`.
            Defaults to `None` for which random selection will be ignored.

        Returns
        -------
        ForecastingModel, Dict
            A tuple containing an untrained 'model_class' instance created from the best-performing hyper-parameters,
            along with a dictionary containing these best hyper-parameters.
        """
        raise_if_not((forecast_horizon is not None) + (val_series is not None) + use_fitted_values == 1,
                     "Please pass exactly one of the arguments 'forecast_horizon', "
                     "'val_target_series' or 'use_fitted_values'.", logger)

        if use_fitted_values:
            raise_if_not(hasattr(model_class(), "fitted_values"),
                         "The model must have a fitted_values attribute to compare with the train TimeSeries",
                         logger)

        elif val_series is not None:
            raise_if_not(series.width == val_series.width,
                         "Training and validation series require the same number of components.",
                         logger)

        # TODO: here too I'd say we can leave these checks to the models
        # if covariates is not None:
        #     raise_if_not(series.has_same_time_as(covariates), 'The provided series and covariates must have the '
        #                                                       'same time axes.')


        # compute all hyperparameter combinations from selection
        params_cross_product = list(product(*parameters.values()))

        #If n_random_samples has been set, randomly select a subset of the full parameter cross product to search with
        if n_random_samples is not None:
            params_cross_product = model_class._sample_params(params_cross_product, n_random_samples)


        # iterate through all combinations of the provided parameters and choose the best one
        iterator = _build_tqdm_iterator(zip(params_cross_product), verbose, total=len(params_cross_product))

        def _evaluate_combination(param_combination):
            param_combination_dict = dict(list(zip(parameters.keys(), param_combination)))
            model = model_class(**param_combination_dict)
            if use_fitted_values:  # fitted value mode
                model._fit_wrapper(series, past_covariates, future_covariates)
                fitted_values = TimeSeries.from_times_and_values(series.time_index, model.fitted_values)
                error = metric(fitted_values, series)
            elif val_series is None:  # expanding window mode
                error = model.backtest(series=series,
                                       past_covariates=past_covariates,
                                       future_covariates=future_covariates,
                                       num_samples=1,
                                       start=start,
                                       forecast_horizon=forecast_horizon,
                                       stride=stride,
                                       metric=metric,
                                       reduction=reduction,
                                       last_points_only=last_points_only)
            else:  # split mode
                model._fit_wrapper(series, past_covariates, future_covariates)
                pred = model._predict_wrapper(len(val_series), series, past_covariates, future_covariates, num_samples=1)
                error = metric(pred, val_series)

            return error

        errors = _parallel_apply(iterator, _evaluate_combination, n_jobs, {}, {})

        min_error = min(errors)

        best_param_combination = dict(list(zip(parameters.keys(), params_cross_product[errors.index(min_error)])))

        logger.info('Chosen parameters: ' + str(best_param_combination))

        return model_class(**best_param_combination), best_param_combination

    def residuals(self,
                  series: TimeSeries,
                  forecast_horizon: int = 1,
                  verbose: bool = False) -> TimeSeries:
        """ A function for computing the residuals produced by the current model on a univariate time series.

        This function computes the difference between the actual observations from `series`
        and the fitted values vector `p` obtained by training the model on `series`.
        For every index `i` in `series`, `p[i]` is computed by training the model on
        `series[:(i - forecast_horizon)]` and forecasting `forecast_horizon` into the future.
        (`p[i]` will be set to the last value of the predicted series.)
        The vector of residuals will be shorter than `series` due to the minimum
        training series length required by the model and the gap introduced by `forecast_horizon`.
        Most commonly, unless otherwise specified, the term "residuals" implies a value for `forecast_horizon` of 1.

        This method works only on univariate series and does not currently support covariates. It uses the median
        prediction (when dealing with stochastic forecasts having num_samples > 1).

        Parameters
        ----------
        series
            The univariate TimeSeries instance which the residuals will be computed for.
        forecast_horizon
            The forecasting horizon used to predict each fitted value.
        verbose
            Whether to print progress.
        Returns
        -------
        TimeSeries
            The vector of residuals.
        """
        series._assert_univariate()

        # get first index not contained in the first training set
        first_index = series.time_index[self.min_train_series_length]

        # compute fitted values
        p = self.historical_forecasts(series=series,
                                      start=first_index,
                                      forecast_horizon=forecast_horizon,
                                      stride=1,
                                      retrain=True,
                                      last_points_only=True,
                                      verbose=verbose)

        # compute residuals
        series_trimmed = series.slice_intersect(p)
        residuals = series_trimmed - (p.quantile_timeseries(quantile=0.5) if p.is_stochastic else p)

        return residuals

    @classmethod
    def _sample_params(model_class, params, n_random_samples):
        """Select the absolute number of samples randomly if an integer has been supplied. If a float has been
        supplied, select a fraction"""

        if isinstance(n_random_samples, int):
            raise_if_not((n_random_samples > 0) and (n_random_samples <= len(params)),
                         "If supplied as an integer, n_random_samples must be greater than 0 and less than or equal to the size of the cartesian product of the hyperparameters.")
            return sample(params, n_random_samples)

        if isinstance(n_random_samples, float):
            raise_if_not((n_random_samples > 0.0) and (n_random_samples <= 1.0),
                         "If supplied as a float, n_random_samples must be greater than 0.0 and less than 1.0.")
            return sample(params, int(n_random_samples * len(params)))

    def _extract_model_creation_params(self):
        """extracts immutable model creation parameters from `ModelMeta` and deletes reference."""
        model_params = copy.deepcopy(self.model_call)
        del self.__class__.model_call
        return model_params

    def untrained_model(self):
        args, kwargs = self._model_params
        return self.__class__(*args, **kwargs)


class GlobalForecastingModel(ForecastingModel, ABC):
    """ The base class for "global" forecasting models, handling several time series and optional covariates.

    Global forecasting models expand upon the functionality of `ForecastingModel` in 4 ways:
    1. Models can be fitted on many series (multivariate or univariate) with different indices.
    2. The input series used by `predict()` can be different from the series used to fit the model.
    3. Covariates can be supported (multivariate or univariate).
    4. They can allow for multivariate target series and covariates.

    The name "global" stems from the fact that a training set of a forecasting model of this class is not constrained
    to a temporally contiguous, "local", time series.

    All implementations have to implement the `fit()` and `predict()` methods defined below.
    The `fit()` method is meant to train the model on one or several training time series, along with optional
    covariates.

    If `fit()` has been called with only one training and covariate series as argument, then calling `predict()` will
    forecast the future of this series. Otherwise, the user has to provide to `predict()` the series they want
    to forecast, as well as covariates, if needed.
    """

    _expect_past_covariates, _expect_future_covariates = False, False
    past_covariate_series, future_covariate_series = None, None

    @abstractmethod
    def fit(self,
            series: Union[TimeSeries, Sequence[TimeSeries]],
            past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
            future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None
            ) -> None:
        """ Fits/trains the model on the provided series

        Defines behavior that should happen when calling the `fit()` method of every global forecasting model.

        Some models support training only on one time series, while others can handle a sequence.
        Similarly, some models can handle past covariates, future covariates, or both.

        Parameters
        ----------
        series
            One or several target time series. The model will be trained to forecast these time series.
            The series may or may not be multivariate, but if multiple series are provided they must have the same number of components.
        past_covariates
            One or several past-observed covariate time series. These time series will not be forecast, but can
            be used by some models as an input. The covariate(s) may or may not be multivariate, but if multiple
            covariates are provided they must have the same number of components.
        future_covariates
            One or several future-known covariate time series. These time series will not be forecast, but can
            be used by some models as an input. The covariate(s) may or may not be multivariate, but if multiple
            covariates are provided they must have the same number of components.
        """

        if isinstance(series, TimeSeries):
            # if only one series is provided, save it for prediction time (including covariates, if available)
            self.training_series = series
            if past_covariates is not None:
                self.past_covariate_series = past_covariates
            if future_covariates is not None:
                self.future_covariate_series = future_covariates
        else:
            if past_covariates is not None:
                self._expect_past_covariates = True
            if future_covariates is not None:
                self._expect_future_covariates = True
        self._fit_called = True

    @abstractmethod
    def predict(self,
                n: int,
                series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
                past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
                future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
                num_samples: int = 1,
                ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """ Forecasts values for a certain number of time steps after the end of the series.

        If `fit()` has been called with only one `TimeSeries` as argument, then the `series` argument of this function
        is optional, and it will simply produce the next `horizon` time steps forecast.
        The `past_covariates` and `future_covariates` arguments also don't have to be provided again in this case.

        If `fit()` has been called with `series` specified as a `Sequence[TimeSeries]`, the `series` argument must
        be specified.

        When the `series` argument is specified, this function will compute the next `n` time steps forecasts
        for the simple series (or for each series in the sequence) given by `series`.

        If multiple past or future covariates were specified during the training, some corresponding covariates must
        also be specified here. For every input in `series` a matching (past and/or future) covariate time series
        has to be provided.

        Parameters
        ----------
        n
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        series
            The series whose future(s) we want to predict.
        past_covariates
            One past-observed covariate time series for every input time series in `series`. They must match the
            past covariates that have been used with the `fit()` function for training in terms of dimension and type.
        future_covariates
            One future-known covariate time series for every input time series in `series`. They must match the
            past covariates that have been used with the `fit()` function for training in terms of dimension and type.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            If `series` is not specified, this function returns a single time series containing the `n`
            next points after then end of the training series.
            If `series` is given and is a simple `TimeSeries`, this function returns the `n` next points
            after the end of `series`.
            If `series` is given and is a sequence of several time series, this function returns
            a sequence where each element contains the corresponding `n` points forecasts.
        """
        if series is None and past_covariates is None and future_covariates is None:
            super().predict(n, num_samples)
        if self._expect_past_covariates and past_covariates is None:
            raise_log(ValueError('The model has been trained with past covariates. Some matching past_covariates '
                                 'have to be provided to `predict()`.'))
        if self._expect_future_covariates and future_covariates is None:
            raise_log(ValueError('The model has been trained with future covariates. Some matching future_covariates '
                                 'have to be provided to `predict()`.'))

    def _predict_wrapper(self, n: int, series: TimeSeries, past_covariates: Optional[TimeSeries],
                         future_covariates: Optional[TimeSeries], num_samples: int) -> TimeSeries:
        return self.predict(n, series, past_covariates=past_covariates,
                            future_covariates=future_covariates, num_samples=num_samples)

    def _fit_wrapper(self, series: TimeSeries, past_covariates: Optional[TimeSeries],
                     future_covariates: Optional[TimeSeries]):
        self.fit(
            series=series,
            past_covariates=past_covariates if self.uses_past_covariates else None,
            future_covariates=future_covariates if self.uses_future_covariates else None
        )

    def _supports_non_retrainable_historical_forecasts(self) -> bool:
        """GlobalForecastingModel supports historical forecasts without retraining the model"""
        return True


class DualCovariatesForecastingModel(ForecastingModel, ABC):
    """ The base class for the forecasting models that are not global, but support future covariates.
    Among other things, it lets Darts forecasting models wrap around statsmodels models
    having a `future_covariates` parameter, which corresponds to future-known covariates.

    All implementations have to implement the `fit()` and `predict()` methods defined below.
    """

    _expect_covariate = False

    def fit(self,
            series: TimeSeries,
            future_covariates: Optional[TimeSeries] = None) -> None:
        """ Fits/trains the model on the provided series.

        Defines behavior that should happen when calling the `fit()` method for the forecasting models handling
        optional future covariates (exogenous variables).

        Parameters
        ----------
        series
            The model will be trained to forecast this time series. Can be multivariate if the model supports it.
        future_covariates
            A time series of future-known covariates. This time series will not be forecasted, but can be used by
            some models as an input. It must contain at least the same time steps/indices as the target `series`.
        """

        if future_covariates is not None:
            if not series.has_same_time_as(future_covariates):
                # fit() expects future_covariates to have same time as the target, so we intersect it here
                future_covariates = future_covariates.slice_intersect(series)

            raise_if_not(series.has_same_time_as(future_covariates),
                         "The provided `future_covariates` series must contain at least the same time steps/"
                         "indices as the target `series`.",
                         logger)
            self._expect_covariate = True

        super().fit(series)

        self._fit(series, future_covariates=future_covariates)

    @abstractmethod
    def _fit(self,
             series: TimeSeries,
             future_covariates: Optional[TimeSeries] = None) -> None:
        """Fits/trains the model on the provided series.
        DualCovariatesModels must implement the fit logic in this method.
        """
        pass

    def predict(self,
                n: int,
                future_covariates: Optional[TimeSeries] = None,
                num_samples: int = 1) -> TimeSeries:
        """ Forecasts values for a certain number of time steps after the end of the series.

        If some future covariates were specified during the training, they must also be specified here.

        Parameters
        ----------
        n
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        future_covariates
            The time series of future-known covariates which can be fed as input to the model. It must correspond to
            the covariate time series that has been used with the `fit()` method for training, and it must contain at
            least the next `n` time steps/indices after the end of the training target series.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.

        Returns
        -------
        TimeSeries, a single time series containing the `n` next points after then end of the training series.
        """

        if future_covariates is None:
            super().predict(n, num_samples)

        if self._expect_covariate and future_covariates is None:
            raise_log(ValueError('The model has been trained with `future_covariates` variable. Some matching '
                                 '`future_covariates` variables have to be provided to `predict()`.'))

        if future_covariates is not None:
            start = self.training_series.end_time() + self.training_series.freq

            invalid_time_span_error = f"For the given forecasting horizon `n={n}`, the provided `future_covariates` " \
                                      f"series must contain at least the next `n={n}` time steps/indices after the " \
                                      f"end of the target `series` that was used to train the model."

            # we raise an error here already to avoid getting error from empty TimeSeries creation
            raise_if_not(future_covariates.end_time() >= start, invalid_time_span_error, logger)

            future_covariates = future_covariates[start:start + (n - 1) * self.training_series.freq]

            raise_if_not(len(future_covariates) == n and self._expect_covariate, invalid_time_span_error, logger)

        return self._predict(n, future_covariates=future_covariates, num_samples=num_samples)

    @abstractmethod
    def _predict(self,
                 n: int,
                 future_covariates: Optional[TimeSeries] = None,
                 num_samples: int = 1) -> TimeSeries:
        """Forecasts values for a certain number of time steps after the end of the series.
        DualCovariatesModels must implement the predict logic in this method.
        """
        pass

    def _fit_wrapper(self, series: TimeSeries, past_covariates: Optional[TimeSeries],
                     future_covariates: Optional[TimeSeries]):
        self.fit(series, future_covariates=future_covariates)

    def _predict_wrapper(self, n: int, series: TimeSeries,
                         past_covariates: Optional[TimeSeries],
                         future_covariates: Optional[TimeSeries],
                         num_samples: int) -> TimeSeries:
        return self.predict(n, future_covariates=future_covariates, num_samples=num_samples)



"""
Models
------
"""

from darts.logging import get_logger

logger = get_logger(__name__)

# Forecasting
from darts.models.forecasting.baselines import NaiveMean, NaiveSeasonal, NaiveDrift
from darts.models.forecasting.exponential_smoothing import ExponentialSmoothing
from darts.models.forecasting.theta import Theta, FourTheta
from darts.models.forecasting.arima import ARIMA
from darts.models.forecasting.fft import FFT
from darts.models.forecasting.varima import VARIMA

try:
    from darts.models.forecasting.auto_arima import AutoARIMA
except ModuleNotFoundError:
    logger.warning("Support for AutoARIMA is not available. To enable it, install u8darts[pmdarima] or u8darts[all].")

try:
    from darts.models.forecasting.prophet_model import Prophet
except ModuleNotFoundError:
    logger.warning('Support for Facebook Prophet is not available. '
                   'To enable it, install "darts", "u8darts[prophet]" or "u8darts[all]" (with pip);'
                   'or "u8darts-all" (with conda).')

try:
    from darts.models.forecasting.block_rnn_model import BlockRNNModel
    from darts.models.forecasting.rnn_model import RNNModel
    from darts.models.forecasting.tcn_model import TCNModel
    from darts.models.forecasting.nbeats import NBEATSModel
    from darts.models.forecasting.transformer_model import TransformerModel
    from darts.models.forecasting.tft_model import TFTModel

except ModuleNotFoundError:
    logger.warning('Support for Torch based models not available. '
                   'To enable them, install "darts", "u8darts[torch]" or "u8darts[all]" (with pip); '
                   'or "u8darts-torch" or "u8darts-all" (with conda).')

# Regression & RegressionEnsemble:
try:
    from darts.models.forecasting.linear_regression_model import LinearRegressionModel
    from darts.models.forecasting.random_forest import RandomForest
    from darts.models.forecasting.regression_model import RegressionModel
    from darts.models.forecasting.regression_ensemble_model import RegressionEnsembleModel
except ModuleNotFoundError:
    logger.warning('Support for Regression based models (incl. RegressionEnsembleModel) not available, '
                   'as they depend on PyTorch Datasets. '
                   'To enable them, install "darts", "u8darts[torch]" or "u8darts[all]" (with pip); '
                   'or "u8darts-torch" or "u8darts-all" (with conda).')

try:
    from darts.models.forecasting.gradient_boosted_model import LightGBMModel
except ModuleNotFoundError:
    logger.warning("Support for LightGBM not available. To enable LightGBM support in Darts, follow the detailed "
                   "install instructions for LightGBM in the README: "
                   "https://github.com/unit8co/darts/blob/master/README.md")

# Ensembling
from darts.models.forecasting.ensemble_model import EnsembleModel
from darts.models.forecasting.baselines import NaiveEnsembleModel

# Filtering
from darts.models.filtering.moving_average import MovingAverage
from darts.models.filtering.kalman_filter import KalmanFilter
from darts.models.filtering.gaussian_process_filter import GaussianProcessFilter



"""
Kalman Filter
-------------
"""

from abc import ABC

from typing import Optional
from filterpy.kalman import KalmanFilter as FpKalmanFilter
from copy import deepcopy
import numpy as np

from darts.models.filtering.filtering_model import FilteringModel
from darts.timeseries import TimeSeries
from darts.utils.utils import raise_if_not


class KalmanFilter(FilteringModel, ABC):
    def __init__(
            self, 
            dim_x: int = 1,
            x_init: Optional[np.array] = None,
            P: Optional[np.array] = None,
            Q: Optional[np.array] = None,
            R: Optional[np.array] = None,
            H: Optional[np.array] = None,
            F: Optional[np.array] = None,
            kf: Optional[FpKalmanFilter] = None
            ):
        """
        This model implements a Kalman filter over a time series (without control signal).

        The key method is `KalmanFilter.filter()`.
        It considers the provided time series as containing (possibly noisy) observations z obtained from a
        (possibly noisy) linear dynamical system with hidden state x. The function `filter(series)` returns a new
        `TimeSeries` describing the distribution of the state x, as inferred by the Kalman filter from
        sequentially observing z from `series`.
        Depending on the use case, this can be used to de-noise a series or infer the underlying hidden state of the
        data generating process (assuming notably that the dynamical system generating the data is known, as captured
        by the `F` matrix.).

        This implementation wraps around filterpy.kalman.KalmanFilter, so more information the parameters can be found
        here: https://filterpy.readthedocs.io/en/latest/kalman/KalmanFilter.html

        The dimensionality of the measurements z is automatically inferred upon calling `filter()`.
        This implementation doesn't include control signal.

        Parameters
        ----------
        dim_x : int
            Size of the Kalman filter state vector. It determines the dimensionality of the `TimeSeries`
            returned by the `filter()` function.
        x_init : ndarray (dim_x, 1), default: [0, 0, ..., 0]
            Initial state; will be updated at each time step.
        P : ndarray (dim_x, dim_x), default: identity matrix
            initial covariance matrix; will be update at each time step
        Q : ndarray (dim_x, dim_x), default: identity matrix
            Process noise covariance matrix
        R : ndarray (dim_z, dim_z), default: identity matrix
            Measurement noise covariance matrix. `dim_z` must match the dimensionality (width) of the `TimeSeries`
            used with `filter()`.
        H : ndarray (dim_z, dim_x), default: all-ones matrix
            measurement function; describes how the measurement z is obtained from the state vector x
        F : ndarray (dim_x, dim_x), default: identity matrix
            State transition matrix; describes how the state evolves from one time step to the next
            in the underlying dynamical system.
        kf : filterpy.kalman.KalmanFilter
            Optionally, an instance of `filterpy.kalman.KalmanFilter`.
            If this is provided, the other parameters are ignored. This instance will be copied for every
            call to `filter()`, so the state is not carried over from one time series to another across several
            calls to `filter()`.
            The various dimensionality in the filter must match those in the `TimeSeries` used when calling `filter()`.
        """
        super().__init__()
        if kf is None:
            self.dim_x = dim_x
            self.x_init = x_init if x_init is not None else np.zeros(self.dim_x,)
            self.P = P if P is not None else np.eye(self.dim_x)
            self.Q = Q if Q is not None else np.eye(self.dim_x)
            self.R = R
            self.H = H
            self.F = F if F is not None else np.eye(self.dim_x)
            self.kf = None
            self.kf_provided = False
        else:
            self.kf = kf
            self.kf_provided = True

    def __str__(self):
        return 'KalmanFilter(dim_x={})'.format(self.dim_x)

    def filter(self,
               series: TimeSeries,
               num_samples: int = 1):
        """
        Sequentially applies the Kalman filter on the provided series of observations.

        Parameters
        ----------
        series : TimeSeries
            The series of observations used to infer the state values according to the specified Kalman process.
            This must be a deterministic series (containing one sample).

        Returns
        -------
        TimeSeries
            A stochastic `TimeSeries` of state values, of dimension `dim_x`.
        """

        raise_if_not(series.is_deterministic, 'The input series for the Kalman filter must be '
                                              'deterministic (observations).')

        dim_z = series.width

        if not self.kf_provided:
            kf = FpKalmanFilter(dim_x=self.dim_x, dim_z=dim_z)
            kf.x = self.x_init
            kf.P = self.P
            kf.Q = self.Q
            kf.R = self.R if self.R is not None else np.eye(dim_z)
            kf.H = self.H if self.H is not None else np.ones((dim_z, self.dim_x))
            kf.F = self.F
        else:
            raise_if_not(dim_z == self.kf.dim_z, 'The provided TimeSeries dimensionality does not match '
                                                 'the filter observation dimensionality dim_z.')
            kf = deepcopy(self.kf)

        super().filter(series)
        values = series.values(copy=False)

        # For each time step, we'll sample "n_samples" from a multivariate Gaussian
        # whose mean vector and covariance matrix come from the Kalman filter.
        if num_samples == 1:
            sampled_states = np.zeros(((len(values)), self.dim_x, ))
        else:
            sampled_states = np.zeros(((len(values)), self.dim_x, num_samples))

        # process_means = np.zeros((len(values), self.dim_x))  # mean values
        # process_covariances = ...                            # covariance matrices; TODO
        for i in range(len(values)):
            obs = values[i, :]
            kf.predict()
            kf.update(obs)
            mean_vec = kf.x.reshape(self.dim_x,)

            if num_samples == 1:
                # It's actually not sampled in this case
                sampled_states[i, :] = mean_vec
            else:
                cov_matrix = kf.P
                sampled_states[i, :, :] = np.random.multivariate_normal(mean_vec, cov_matrix, size=num_samples).T

        # TODO: later on for a forecasting model we'll have to do something like
        """
        for _ in range(horizon):
            kf.predict()
            # forecasts on the observations, obtained from the state
            preds.append(kf.H.dot(kf.x))
            preds_cov.append(kf.H.dot(kf.P).dot(kf.H.T))
        """

        return TimeSeries.from_times_and_values(series.time_index, sampled_states)



"""
Model selection utilities
-------------------------
Utilities that help in model selection e.g. by splitting a dataset.
"""

from typing import Sequence, Optional, Union, Tuple
from darts import TimeSeries

MODEL_AWARE = 'model-aware'
SIMPLE = 'simple'


class SplitTimeSeriesSequence(Sequence):
    """
    This class is primarily meant to be instantiated from ``train_test_split()`` function.
    """
    def __init__(self,
                 type: str,
                 data: Sequence[TimeSeries],
                 test_size: Optional[Union[float, int]] = 0.25,
                 axis: Optional[int] = 0,
                 input_size: Optional[int] = None,
                 horizon: Optional[int] = None,
                 vertical_split_type: Optional[str] = SIMPLE):

        if type not in ['train', 'test']:
            raise AttributeError('Value for type parameter should be either `train` or `test`')

        if not data:
            raise AttributeError('The `data` parameter cannot be empty.')

        if axis not in [0, 1]:
            raise AttributeError('The `axis` parameter should be either 0 or 1.')

        if vertical_split_type not in [SIMPLE, MODEL_AWARE]:
            raise AttributeError('`vertical_split_type` can be eiter `simple` or `model-aware`.')

        if axis == 1 and vertical_split_type == MODEL_AWARE and (horizon == 0 or input_size == 0):
                raise AttributeError("You need to provide non-zero `horizon` and `input_size` parameters when axis=1")

        self.type = type
        self.data = data
        self.test_size = test_size
        self.axis = axis
        self.input_size = input_size
        self.horizon = horizon
        self.vertical_split_type = vertical_split_type
        self._horizontal_split_index = None

    def _get_horizontal_split_index(self):
        if not self._horizontal_split_index:
            if 0 < self.test_size < 1:
                self._horizontal_split_index = int(len(self.data) * (1 - self.test_size))
            else:
                self._horizontal_split_index = len(self.data) - self.test_size

        return self._horizontal_split_index

    def _get_vertical_split_indices(self, ts_length):

        if self.vertical_split_type == SIMPLE:
            if 0 < self.test_size < 1:
                test_size = int(ts_length * self.test_size)
            else:
                test_size = self.test_size

            test_start_index = ts_length - test_size
            train_end_index = test_start_index

            if test_start_index < 0:
                raise AttributeError("`test_size` is bigger then timeseries length")

        else: # model-aware split

            # an example on how these are calculated:
            # test_size = 4
            # input_size = 5 (-)
            # horizon = 3 (+)
            # len(data) = 16 (*)

            # 0         5         10          15 <- index
            # * * * * * * * * * * * * * * * *
            #                 - - - - - + + +   < test sample 1
            #               - - - - - + + +     < test sample 2
            #             - - - - - + + +       < test sample 3
            #           - - - - - + + +         < test sample 4
            #           ^         ^
            #           |         train_end_index = 10 (note the [:train_end_index] slice stops at 9)
            #           test_start_index = 5

            if 0 < self.test_size < 1:
                test_size = int((ts_length - self.horizon) * self.test_size)
            else:
                test_size = self.test_size

            train_end_index = ts_length - self.horizon - test_size + 1
            test_start_index = ts_length - self.horizon - self.input_size - test_size + 1

            if train_end_index < 0:
                train_end_index = 0

            if train_end_index < self.input_size + self.horizon:
                # Note: note that we don't have to check for test_start_index < 0 because train_end_index will always
                # raise exception first.
                raise AttributeError("Not enough data to create training and test sets")

        return train_end_index, test_start_index

    def __getitem__(self, i: int) -> TimeSeries:
        if self.axis == 0:
            split_index = self._get_horizontal_split_index()
            if self.type == 'train':
                if i >= split_index:
                    raise IndexError('Exceeded the size of the training sequence.')
                return self.data[i]
            else:
                if i + split_index >= len(self.data):
                    raise IndexError('Exceeded the size of the test sequence.')
                return self.data[split_index + i]
        else: # axis == 1
            train_end_index, test_start_index = self._get_vertical_split_indices(len(self.data[i]))
            if self.type == 'train':
                return self.data[i][:train_end_index]
            else:
                return self.data[i][test_start_index:]

    def __len__(self):
        if self.axis == 0:
            split_index = self._get_horizontal_split_index()
            if self.type == 'train':
                return split_index
            else:
                return len(self.data) - split_index
        else:
            return len(self.data)

    @classmethod
    def make_splitter(cls,
                      data: Union[TimeSeries, Sequence[TimeSeries]],
                      test_size: Optional[Union[float, int]] = 0.25,
                      axis: Optional[int] = 0,
                      input_size: Optional[int] = 0,
                      horizon: Optional[int] = 0,
                      vertical_split_type: Optional[str] = SIMPLE,
                      lazy: bool = False
                      ) -> Union[Tuple[TimeSeries, TimeSeries], Tuple[Sequence[TimeSeries], Sequence[TimeSeries]]]:

        if not isinstance(data, Sequence):
            axis = 1
            data = [data]  # convert to sequence for unified processing later
            single_timeseries = True
        else:
            single_timeseries = False

        train_set = cls(type='train', data=data, test_size=test_size, axis=axis, input_size=input_size, horizon=horizon,
                        vertical_split_type=vertical_split_type)

        test_set = cls(type='test', data=data, test_size=test_size, axis=axis, input_size=input_size, horizon=horizon,
                       vertical_split_type=vertical_split_type)

        if single_timeseries:
            return train_set[0], test_set[0]
        else:
            if lazy:
                return train_set, test_set
            else:
                return list(train_set), list(test_set)


def train_test_split(
        data: Union[TimeSeries, Sequence[TimeSeries]],
        test_size: Optional[Union[float, int]] = 0.25,
        axis: Optional[int] = 0,
        input_size: Optional[int] = 0,
        horizon: Optional[int] = 0,
        vertical_split_type: Optional[str] = SIMPLE,
        lazy: bool = False
        ) -> Union[Tuple[TimeSeries, TimeSeries], Tuple[Sequence[TimeSeries], Sequence[TimeSeries]]]:
    """
    Splits the dataset into training and test dataset. Supports splitting along the sample axis and time axis.

    If the input type is single TimeSeries, then only splitting over time axis is available, thus ``input_size`` and
    ``horizon`` have to be provided.

    When splitting over the time axis, splitter tries to greedy satisfy the requested test set size, i.e. when one of
    the timeseries in sequence is too small, all samples will go to the test set and the exception will be raised.


    Time axis split with ``model-aware`` split type enabled tries to reclaim as much datapoints for training as
    possible by partially overlapping test set with training set. This is possible because only forecasted part of
    test set cannot be used for training. The formula to calculate last available timestep for training set is
    following:

        train end index = ts_length - self.horizon - test_size

    And the formula to calculate the first timestep of test dataset is following:

        test start index = timeseries length - horizon - input_size - test_size + 1

    Parameters
    ----------
    data
        original dataset to split into training and test

    test_size
        size of the test set. If the value is between 0 and 1, parameter is treated as a split proportion. Otherwise
        it is treated as a absolute number of samples from each timeseries that will be in the test set. [default = 0.25]

    axis
        Axis to split the dataset on. When 0 (default) it is split on samples. Otherwise, if ``axis = 1``,
        timeseries are split along time axis. Note that for single timeseries the default option is 1 (0 makes
        no sense). [default: 0 for sequence of timeseries, 1 for timeseries]

    input_size
        size of the input. Only valid with ``vertical_split_type == 'model-aware'``. [default: None]

    horizon
        forecast horizon. Only valid with ``vertical_split_type == 'model-aware'``. [default: None]

    vertical_split_type
        can be either ``simple``, where the exact number from test size will be deducted from timeseries for test set and
        remaining will go to training set; or ``model-aware``, where you have to provide ``input_size`` and ``horizon``
        as well. Note, that second option is more efficient timestep-wise, since training and test sets will be
        partially overlapping. [default: ``simple``]

    lazy
        by default, train and test datasets are returned as a list of timeseries. This may be memory
        inefficient if dataset is large, so setting this flag allows instead to return a ``Sequence`` object
        loading the data lazily. Warning: turning ``lazy`` on disables some sanity checks for the datasets
        that may result in exceptions during sample generation. [default: False]

    Returns
    -------
    tuple of two Sequence[TimeSeries], or tuple of two Timeseries
        Training and test datasets tuple.
    """

    return SplitTimeSeriesSequence.make_splitter(data, test_size, axis, input_size, horizon, vertical_split_type, lazy)



"""
Encoder Classes Main
--------------------

Encoders can generate past and/or future covariate series by encoding the index of a TimeSeries `series`.
Each encoder class has an `encode_train()` and `encode_inference()` to generate the encodings for training and
inference.

The encoders extract the index either from the target series or optional additional past/future covariates.
If additional covariates are supplied to `encode_train()` or `encode_inference()`, the time index of those
covariates are used for the encodings. This means that the input covariates must meet the same model-specific
requirements as wihtout encoders.

There are two main types of encoder classes: `SingleEncoder` and `SequentialEncoder`.

*   SingleEncoder
        The SingleEncoder classes carry the encoder logic for past and future covariates, and training and
        inference datasets. They can be used as stand-alone encoders.

        Each SingleEncoder has a dedicated subclass for generating past or future covariates. The naming convention
        is `{X}{SingleEncoder}` where {X} is one of (Past, Future) and {SingleEncoder} is one of the SingleEncoder
        classes described in the next section. An example:

        .. highlight:: python
        .. code-block:: python

            encoder = PastDatetimeAttributeEncoder(input_chunk_length=24, output_chunk_length=12, attribute='month')

            past_covariates_train = encoder.encode_train(target=target, covariate=optional_past_covariates)
            past_covariates_inf = encoder.encode_inference(n=12, target=target, covariate=optional_past_covariates)

*   SequentialEncoder
        Stores and controls multiple SingleEncoders for both past and/or future covariates all under one hood.
        It provides the same functionality as SingleEncoders (`encode_train()` and `encode_inference()`).
        It can be used both as stand-alone or as an all-in-one solution with Darts' `TorchForecastingModel` models
        through optional parameter `add_encoders`:

        .. highlight:: python
        .. code-block:: python

            model = SomeTorchForecastingModel(..., add_encoders={...})
        ..

        If used at model creation, the SequentialEncoder will handle all past and future encoders autonomously.
        The requirements for model parameter `add_encoders` are described in the next section or in
        :meth:`SequentialEncoder <darts.utils.data.encoders.SequentialEncoder>`.

SingleEncoder
-------------

The SingleEncoders from {X}{SingleEncoder} are:

*   DatetimeAttributeEncoder
        Adds scalar pd.DatatimeIndex attribute information derived from `series.time_index`.
        Requires `series` to have a pd.DatetimeIndex.

        attribute
            An attribute of `pd.DatetimeIndex`: see all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
*   CyclicTemporalEncoder
        Adds cyclic pd.DatetimeIndex attribute information deriveed from `series.time_index`.
        Adds 2 columns, corresponding to sin and cos encodings, to uniquely describe the underlying attribute.
        Requires `series` to have a pd.DatetimeIndex.

        attribute
            An attribute of `pd.DatetimeIndex` that follows a cyclic pattern. One of ('month', 'day', 'weekday',
            'dayofweek', 'day_of_week', 'hour', 'minute', 'second', 'microsecond', 'nanosecond', 'quarter',
            'dayofyear', 'day_of_year', 'week', 'weekofyear', 'week_of_year').
*   IntegerIndexEncoder
        Adds absolute or relative index positions as integer values (positions) derived from `series` time index.
        `series` can either have a pd.DatetimeIndex or an integer index.

        attribute
            Either 'absolute' or 'relative'.
            'absolute' will generate position values ranging from 0 to inf where 0 is set at the start of `series`.
            'relative' will generate position values relative to the forecasting/prediction point. Values range
            from -inf to inf where 0 is set at the forecasting point.
*   CallableIndexEncoder
        Applies a user-defined callable to encode `series`' index.
        `series` can either have a pd.DatetimeIndex or an integer index.

        attribute
            a callable/function to encode the index.
            For `series` with a pd.DatetimeIndex: `lambda index: (index.year - 1950) / 50`
            For `series` with an integer index:`lambda index: index / 50`

SequentialEncoder
-----------------

The SequentialEncoder combines the logic of all SingleEncoders from above and has additional benefits:

*   use multiple encoders at once
*   generate multiple attribute encodings at once
*   generate both past and future at once
*   supports transformers (Scaler)
*   easy to use with TorchForecastingModels

The model parameter `add_encoders` must be a Dict following of this convention:

*   outer keys: `SingleEncoder` and Transformer tags:

    *   'datetime_attribute' for `DatetimeAttributeEncoder`
    *   'cyclic' for `CyclicEncoder`
    *   'position' for `IntegerIndexEncoder`
    *   'custom' for `CallableIndexEncoder`
    *   'transformer' for a transformer
*   inner keys: covariate type

    *   'past' for past covariates
    *   'future' for future covariates
    *   (do not specify for 'transformer')
*   inner key values:

    *   list of attributes for `SingleEncoder`
    *   transformer object for 'transformer'

Below is an example that illustrates a valid `add_encoders` dict for hourly data and how it can be used with a
TorchForecastingModel (this is only meant to illustrate many features at once).

.. highlight:: python
.. code-block:: python

    add_encoders = {
        'cyclic': {'future': ['month']},
        'datetime_attribute': {'past': ['hour'], 'future': ['year', 'dayofweek']},
        'position': {'past': ['absolute'], 'future': ['relative']},
        'custom': {'past': [lambda index: (index.year - 1950) / 50]},
        'transformer': Scaler()
    }

    model = SomeTorchForecastingModel(..., add_encoders=add_encoders)
..
"""

import pandas as pd
import numpy as np
import copy

from typing import Union, Optional, Dict, List, Sequence, Tuple, Callable

from darts import TimeSeries
from darts.timeseries import DIMS
from darts.utils.data.encoder_base import (ReferenceIndexType,
                                           CovariateIndexGenerator,
                                           PastCovariateIndexGenerator,
                                           FutureCovariateIndexGenerator,
                                           Encoder,
                                           SingleEncoder,
                                           SequentialEncoderTransformer,
                                           SupportedIndex)
from darts.utils.data.utils import _index_diff

from darts.logging import raise_if_not, get_logger, raise_if

from darts import concatenate
from darts.utils.timeseries_generation import datetime_attribute_timeseries
from darts.dataprocessing.transformers import FittableDataTransformer

SupportedTimeSeries = Union[TimeSeries, Sequence[TimeSeries]]
logger = get_logger(__name__)

ENCODER_KEYS = ['cyclic', 'datetime_attribute', 'position', 'custom']
FUTURE = 'future'
PAST = 'past'
VALID_TIME_PARAMS = [FUTURE, PAST]
VALID_ENCODER_DTYPES = (str, Sequence)

TRANSFORMER_KEYS = ['transformer']
VALID_TRANSFORMER_DTYPES = FittableDataTransformer
INTEGER_INDEX_ATTRIBUTES = ['absolute', 'relative']


class CyclicTemporalEncoder(SingleEncoder):
    """CyclicTemporalEncoder: Cyclic index encoding for `TimeSeries` that have a time index of type `pd.DatetimeIndex`.
    """

    def __init__(self, index_generator: CovariateIndexGenerator, attribute: str):
        """
        Parameters
        ----------
        index_generator
            An instance of `CovariateIndexGenerator` with methods `generate_train_series()` and
            `generate_inference_series()`. Used to generate the index for encoders.
        attribute
            The attribute of the underlying pd.DatetimeIndex from  for which to apply cyclic encoding.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        """
        super(CyclicTemporalEncoder, self).__init__(index_generator)
        self.attribute = attribute

    def _encode(self, index: SupportedIndex, dtype: np.dtype) -> TimeSeries:
        """applies cyclic encoding from `datetime_attribute_timeseries()` to `self.attribute` of `index`.
        """
        super(CyclicTemporalEncoder, self)._encode(index, dtype)
        return datetime_attribute_timeseries(index, attribute=self.attribute, cyclic=True, dtype=dtype)

    @property
    def accept_transformer(self) -> List[bool]:
        """CyclicTemporalEncoder should not be transformed. Returns two elements for sine and cosine waves."""
        return [False, False]


class PastCyclicEncoder(CyclicTemporalEncoder):
    """Cyclic encoder for past covariates."""

    def __init__(self, input_chunk_length, output_chunk_length, attribute):
        """
        Parameters
        ----------
        input_chunk_length
            The length of the emitted past series.
        output_chunk_length
            The length of the emitted future series.
        attribute
            The attribute of the underlying pd.DatetimeIndex from  for which to apply cyclic encoding.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        """
        super(PastCyclicEncoder, self).__init__(
            index_generator=PastCovariateIndexGenerator(input_chunk_length, output_chunk_length),
            attribute=attribute
        )


class FutureCyclicEncoder(CyclicTemporalEncoder):
    """Cyclic encoder for future covariates."""

    def __init__(self, input_chunk_length, output_chunk_length, attribute):
        """
        Parameters
        ----------
        input_chunk_length
            The length of the emitted past series.
        output_chunk_length
            The length of the emitted future series.
        attribute
            The attribute of the underlying pd.DatetimeIndex from  for which to apply cyclic encoding.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        """
        super(FutureCyclicEncoder, self).__init__(
            index_generator=FutureCovariateIndexGenerator(input_chunk_length, output_chunk_length),
            attribute=attribute
        )


class DatetimeAttributeEncoder(SingleEncoder):
    """DatetimeAttributeEncoder: Adds pd.DatatimeIndex attribute information derived from the index as scalars.
    Requires the underlying TimeSeries to have a pd.DatetimeIndex
    """

    def __init__(self, index_generator: CovariateIndexGenerator, attribute: str):
        """
        Parameters
        ----------
        index_generator
            An instance of `CovariateIndexGenerator` with methods `generate_train_series()` and
            `generate_inference_series()`. Used to generate the index for encoders.
        attribute
            The attribute of the underlying pd.DatetimeIndex from  for which to add scalar information.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        """
        super(DatetimeAttributeEncoder, self).__init__(index_generator)
        self.attribute = attribute

    def _encode(self, index: SupportedIndex, dtype: np.dtype) -> TimeSeries:
        """Applies cyclic encoding from `datetime_attribute_timeseries()` to `self.attribute` of `index`.
        """
        super(DatetimeAttributeEncoder, self)._encode(index, dtype)
        return datetime_attribute_timeseries(index, attribute=self.attribute, dtype=dtype)

    @property
    def accept_transformer(self) -> List[bool]:
        """DatetimeAttributeEncoder accepts transformations"""
        return [True]


class PastDatetimeAttributeEncoder(DatetimeAttributeEncoder):
    """Datetime attribute encoder for past covariates."""

    def __init__(self, input_chunk_length, output_chunk_length, attribute):
        """
        Parameters
        ----------
        input_chunk_length
            The length of the emitted past series.
        output_chunk_length
            The length of the emitted future series.
        attribute
            The attribute of the underlying pd.DatetimeIndex from  for which to add scalar information.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        """
        super(PastDatetimeAttributeEncoder, self).__init__(
            index_generator=PastCovariateIndexGenerator(input_chunk_length, output_chunk_length),
            attribute=attribute
        )


class FutureDatetimeAttributeEncoder(DatetimeAttributeEncoder):
    """Datetime attribute encoder for future covariates."""

    def __init__(self, input_chunk_length, output_chunk_length, attribute):
        """
        Parameters
        ----------
        input_chunk_length
            The length of the emitted past series.
        output_chunk_length
            The length of the emitted future series.
        attribute
            The attribute of the underlying pd.DatetimeIndex from  for which to add scalar information.
            Must be an attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month",
            "weekday", "day", "hour", "minute", "second". See all available attributes in
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex.
            For more information, check out :meth:`datetime_attribute_timeseries()
            <darts.utils.timeseries_generation.datetime_attribute_timeseries>`
        """
        super(FutureDatetimeAttributeEncoder, self).__init__(
            index_generator=FutureCovariateIndexGenerator(input_chunk_length, output_chunk_length),
            attribute=attribute
        )


class IntegerIndexEncoder(SingleEncoder):
    """IntegerIndexEncoder: Adds integer index value (position) derived from the underlying TimeSeries' time index
    for past and future covariates.
    """

    def __init__(self, index_generator: CovariateIndexGenerator, attribute: str):
        """
        Parameters
        ----------
        index_generator
            An instance of `CovariateIndexGenerator` with methods `generate_train_series()` and
            `generate_inference_series()`. Used to generate the index for encoders.
        attribute
            Either 'absolute' or 'relative'. If 'absolute', the generated encoded values will range from (0, inf)
            and the train target series will be used as a reference to set the 0-index. If 'relative', the generated
            encoded values will range from (-inf, inf) and the train target series end time will be used as a reference
            to evaluate the relative index positions.
        """
        raise_if_not(isinstance(attribute, str) and attribute in INTEGER_INDEX_ATTRIBUTES,
                     f'Encountered invalid encoder argument `{attribute}` for encoder `position`. '
                     f'Attribute must be one of `("absolute", "relative")`.',
                     logger)

        super(IntegerIndexEncoder, self).__init__(index_generator)

        self.attribute = attribute
        self.reference_index: Optional[Tuple[int, Optional[Union[pd.Timestamp, int]]]] = None
        self.was_called = False

    def _encode(self, index: SupportedIndex, dtype: np.dtype) -> TimeSeries:
        """Applies cyclic encoding from `datetime_attribute_timeseries()` to `self.attribute` of `index`.
        1)  for attribute=='absolute', the reference point/index is one step before start of the train target series
        2)  for attribute=='relative', the reference point/index is the overall prediction/forecast index
        """
        super(IntegerIndexEncoder, self)._encode(index, dtype)

        # load reference index from index_generators
        if not self.was_called:
            self.reference_index = self.index_generator.reference_index
            self.was_called = True

        current_start_value = index[0]

        # extract reference index
        reference_index, reference_value = self.reference_index

        # get the difference between last index and reference index for each case
        index_diff = _index_diff(self=current_start_value, other=reference_value, freq=index.freq)
        # set the start integer index value for the current index
        current_start_index = reference_index - index_diff if self.attribute == 'absolute' else - index_diff

        encoded = TimeSeries.from_times_and_values(times=index,
                                                   values=np.arange(current_start_index,
                                                                    current_start_index + len(index)),
                                                   columns=[self.attribute + '_idx']).astype(np.dtype(dtype))

        # update reference index for 'absolute' case to avoid having to evaluate longer differences (cost-intensive)
        if self.attribute == 'absolute':
            self.reference_index = (current_start_index + len(encoded) - 1, encoded.time_index[-1])
        return encoded

    @property
    def accept_transformer(self) -> List[bool]:
        """IntegerIndexEncoder accepts transformations. Note that transforming 'relative' IntegerIndexEncoder
        will return an 'absolute' index."""
        return [True]


class PastIntegerIndexEncoder(IntegerIndexEncoder):
    """IntegerIndexEncoder: Adds integer index value (position) for past covariates derived from the underlying
    TimeSeries' time index.
    """

    def __init__(self, input_chunk_length: int, output_chunk_length: int, attribute: str):
        """
        Parameters
        ----------
        input_chunk_length
            The length of the emitted past series.
        output_chunk_length
            The length of the emitted future series.
        attribute
            Either 'absolute' or 'relative'. If 'absolute', the generated encoded values will range from (0, inf)
            and the train target series will be used as a reference to set the 0-index. If 'relative', the generated
            encoded values will range from (-inf, inf) and the train target series end time will be used as a reference
            to evaluate the relative index positions.
        """
        reference_index_type = ReferenceIndexType.PREDICTION if attribute == 'relative' else ReferenceIndexType.START

        super(PastIntegerIndexEncoder, self).__init__(
            index_generator=PastCovariateIndexGenerator(input_chunk_length,
                                                        output_chunk_length,
                                                        reference_index_type=reference_index_type),
            attribute=attribute
        )


class FutureIntegerIndexEncoder(IntegerIndexEncoder):
    """IntegerIndexEncoder: Adds integer index value (position) for future covariates derived from the underlying
    TimeSeries' time index.
    """

    def __init__(self, input_chunk_length: int, output_chunk_length: int, attribute: str):
        """
        Parameters
        ----------
        input_chunk_length
            The length of the emitted past series.
        output_chunk_length
            The length of the emitted future series.
        attribute
            Either 'absolute' or 'relative'. If 'absolute', the generated encoded values will range from (0, inf)
            and the train target series will be used as a reference to set the 0-index. If 'relative', the generated
            encoded values will range from (-inf, inf) and the train target series end time will be used as a reference
            to evaluate the relative index positions.
        """
        reference_index_type = ReferenceIndexType.PREDICTION if attribute == 'relative' else ReferenceIndexType.START

        super(FutureIntegerIndexEncoder, self).__init__(
            index_generator=FutureCovariateIndexGenerator(input_chunk_length,
                                                          output_chunk_length,
                                                          reference_index_type=reference_index_type),
            attribute=attribute
        )


class CallableIndexEncoder(SingleEncoder):
    """CallableIndexEncoder: Applies a user-defined callable to encode the underlying index for past and future
    covariates.
    """

    def __init__(self, index_generator: CovariateIndexGenerator, attribute: Callable):
        """
        Parameters
        ----------
        index_generator
            An instance of `CovariateIndexGenerator` with methods `generate_train_series()` and
            `generate_inference_series()`. Used to generate the index for encoders.
        attribute
            A callable that takes an index `index` of type `(pd.DatetimeIndex, pd.Int64Index, pd.RangeIndex)` as input
            and returns a np.ndarray of shape `(len(index),)`.
            An example for a correct `attribute` for `ìndex` of type pd.DatetimeIndex:
                `attribute = lambda index: (index.year - 1950) / 50]}`
            And for pd.Int64Index or pd.RangeIndex:
                `attribute = lambda index: (index - 1950) / 50]}`
        """
        raise_if_not(callable(attribute),
                     f'Encountered invalid encoder argument `{attribute}` for encoder `callable`. '
                     f'Attribute must be a callable that returns a `np.ndarray`.',
                     logger)

        super(CallableIndexEncoder, self).__init__(index_generator)

        self.attribute = attribute

    def _encode(self, index: SupportedIndex, dtype: np.dtype) -> TimeSeries:
        """Apply the user-defined callable to encode the index
        """
        super(CallableIndexEncoder, self)._encode(index, dtype)

        return TimeSeries.from_times_and_values(times=index,
                                                values=self.attribute(index),
                                                columns=['custom']).astype(np.dtype(dtype))

    @property
    def accept_transformer(self) -> List[bool]:
        """CallableIndexEncoder accepts transformations."""
        return [True]


class PastCallableIndexEncoder(CallableIndexEncoder):
    """IntegerIndexEncoder: Adds integer index value (position) for past covariates derived from the underlying
    TimeSeries' time index.
    """

    def __init__(self, input_chunk_length: int, output_chunk_length: int, attribute: Union[str, Callable]):
        """
        Parameters
        ----------
        input_chunk_length
            The length of the emitted past series.
        output_chunk_length
            The length of the emitted future series.
        attribute
            A callable that takes an index `index` of type `(pd.DatetimeIndex, pd.Int64Index, pd.RangeIndex)` as input
            and returns a np.ndarray of shape `(len(index),)`.
            An example for a correct `attribute` for `ìndex` of type pd.DatetimeIndex:
                `attribute = lambda index: (index.year - 1950) / 50}`
            And for pd.Int64Index or pd.RangeIndex:
                `attribute = lambda index: (index - 1950) / 50}`
        """
        super(PastCallableIndexEncoder, self).__init__(
            index_generator=PastCovariateIndexGenerator(input_chunk_length, output_chunk_length),
            attribute=attribute
        )


class FutureCallableIndexEncoder(CallableIndexEncoder):
    """IntegerIndexEncoder: Adds integer index value (position) for future covariates derived from the underlying
    TimeSeries' time index.
    """

    def __init__(self, input_chunk_length: int, output_chunk_length: int, attribute: Union[str, Callable]):
        """
        Parameters
        ----------
        input_chunk_length
            The length of the emitted past series.
        output_chunk_length
            The length of the emitted future series.
        attribute
            A callable that takes an index `index` of type `(pd.DatetimeIndex, pd.Int64Index, pd.RangeIndex)` as input
            and returns a np.ndarray of shape `(len(index),)`.
            An example for a correct `attribute` for `ìndex` of type pd.DatetimeIndex:
                `attribute = lambda index: (index.year - 1950) / 50]}`
            And for pd.Int64Index or pd.RangeIndex:
                `attribute = lambda index: (index - 1950) / 50]}`
        """

        super(FutureCallableIndexEncoder, self).__init__(
            index_generator=FutureCovariateIndexGenerator(input_chunk_length, output_chunk_length),
            attribute=attribute
        )


class SequentialEncoder(Encoder):
    """A `SequentialEncoder` object can store and control multiple past and future covariate encoders at once.
    It provides the same functionality as single encoders (`encode_train()` and `encode_inference()`).
    """

    def __init__(self,
                 add_encoders: Dict,
                 input_chunk_length: int,
                 output_chunk_length: int,
                 takes_past_covariates: bool = False,
                 takes_future_covariates: bool = False) -> None:

        """
        SequentialEncoder automatically creates encoder objects from parameter `add_encoders` used when creating a
        `TorchForecastingModel`.

        *   Only kwarg `add_encoders` of type `Optional[Dict]` will be used to extract the encoders.
            For example: `model = MyModel(..., add_encoders={...}, ...)`

        The `add_encoders` dict must follow this convention:
            `{encoder keyword: {temporal keyword: List[attributes]}, ..., transformer keyword: transformer object}`
        Supported encoder keywords:
            `'cyclic'` for cyclic temporal encoder. See the docs
            :meth:`CyclicTemporalEncoder <darts.utils.data.encoders.CyclicTemporalEncoder>`;
            `'datetime_attribute'` for adding scalar information of pd.DatetimeIndex attribute. See the docs
            :meth:`DatetimeAttributeEncoder <darts.utils.data.encoders.DatetimeAttributeEncoder>`
            `'position'` for integer index position encoder. See the docs
            :meth:`IntegerIndexEncoder <darts.utils.data.encoders.IntegerIndexEncoder>`;
            `'custom'` for encoding index with custom callables (functions). See the docs
            :meth:`CallableIndexEncoder <darts.utils.data.encoders.CallableIndexEncoder>`;
        Supported temporal keywords:
            'past' for adding encoding as past covariates
            'future' for adding encoding as future covariates
        Supported attributes:
            for attributes read the referred docs for the corresponding encoder from above
        Supported transformers:
            a transformer can be added with transformer keyword 'transformer'. The transformer object must be an
            instance of Darts' :meth:`FittableDataTransformer
            <darts.dataprocessing.transformers.fittable_data_transformer.FittableDataTransformer>` such as Scaler() or
            BoxCox(). The transformers will be fitted on the training dataset when calling calling `model.fit()`.
            The training, validation and inference datasets are then transformed equally.

        An example of a valid `add_encoders` dict for hourly data:
            from darts.dataprocessing.transformers import Scaler
            add_encoders={
                'cyclic': {'future': ['month']},
                'datetime_attribute': {'past': ['hour'], 'future': ['year', 'dayofweek']},
                'position': {'past': ['absolute'], 'future': ['relative']},
                'custom': {'past': [lambda index: (index.year - 1950) / 50]},
                'transformer': Scaler()
            }

        Tuples of `(encoder_id, attribute)` are extracted from `add_encoders` to instantiate the `SingleEncoder`
        objects:
        * The `encoder_id` is extracted as follows:
            str(encoder_kw) + str(temporal_kw) -> 'cyclic' + 'past' -> `encoder_id` = 'cyclic_past'
            The `encoder_id` is used to map the parameters with the corresponding `SingleEncoder` objects.
        * The `attribute` is extracted from the values given by values under `temporal_kw`
            `attribute` = 'month'
            ...
            The `attribute` tells the `SingleEncoder` which attribute of the index to encode

        New encoders can be added by appending them to the mapping property `SequentialEncoder.encoder_map()`

        Parameters
        ----------
        add_encoders
            The parameters used at `TorchForecastingModel` model creation.
        input_chunk_length
            The length of the emitted past series.
        output_chunk_length
            The length of the emitted future series.
        takes_past_covariates
            Whether or not the `TrainingDataset` takes past covariates
        takes_future_covariates
            Whether or not the `TrainingDataset` takes past covariates
        """

        super(SequentialEncoder, self).__init__()
        self.params = add_encoders
        self.input_chunk_length = input_chunk_length
        self.output_chunk_length = output_chunk_length
        self.train_called = False
        self.encoding_available = False
        self.takes_past_covariates = takes_past_covariates
        self.takes_future_covariates = takes_future_covariates

        # encoders
        self._past_encoders: List[SingleEncoder] = []
        self._future_encoders: List[SingleEncoder] = []

        # transformer
        self._past_transformer: Optional[SequentialEncoderTransformer] = None
        self._future_transformer: Optional[SequentialEncoderTransformer] = None

        # setup encoders and transformer
        self._setup_encoders(self.params)
        self._setup_transformer(self.params)

    def encode_train(self,
                     target: SupportedTimeSeries,
                     past_covariate: Optional[SupportedTimeSeries] = None,
                     future_covariate: Optional[SupportedTimeSeries] = None,
                     encode_past: bool = True,
                     encode_future: bool = True) -> Tuple[Sequence[TimeSeries], Sequence[TimeSeries]]:
        """Returns encoded index for all past and/or future covariates for training.
        Which covariates are generated depends on the parameters used at model creation.
        
        Parameters
        ----------
        target
            The target TimeSeries used during training or passed to prediction as `series`
        past_covariate
            Optionally, the past covariates used for training.
        future_covariate
            Optionally, the future covariates used for training.
        encode_past
            Whether or not to apply encoders for past covariates
        encode_future
            Whether or not to apply encoders for future covariates
        Returns
        -------
        Tuple[past_covariate, future_covariate]
            The past_covariate and/or future_covariate for training including the encodings.
            If input {x}_covariate is None and no {x}_encoders are given, will return `None`
            for the {x}_covariate.
        Raises
        ------
        Warning
            If model was created with `add_encoders` and there is suspicion of lazy loading.
            The encodings/covariates are generated eagerly before starting training for all individual targets and
            loaded into memory. Depending on the size of target data, this can create memory
            issues. In case this applies, consider setting `add_encoders=None` at model
            creation and build your encodings covariates manually for lazy loading.
        """
        if not self.train_called:
            if not isinstance(target, (TimeSeries, list)):
                logger.warning("Fitting was called with `add_encoders` and suspicion of lazy loading. "
                               "The encodings/covariates are generated pre-train for all individual targets and "
                               "loaded into memory. Depending on the size of your data, this can create memory issues. "
                               "In case this applies, consider setting `add_encoders=None` at model creation.")

            self.train_called = True

        return self._launch_encoder(target=target,
                                    past_covariate=past_covariate,
                                    future_covariate=future_covariate,
                                    n=None,
                                    encode_past=encode_past,
                                    encode_future=encode_future)

    def encode_inference(self,
                         n: int,
                         target: SupportedTimeSeries,
                         past_covariate: Optional[SupportedTimeSeries] = None,
                         future_covariate: Optional[SupportedTimeSeries] = None,
                         encode_past: bool = True,
                         encode_future: bool = True) -> Tuple[Sequence[TimeSeries], Sequence[TimeSeries]]:
        """Returns encoded index for all past and/or future covariates for inference/prediction.
        Which covariates are generated depends on the parameters used at model creation.

        Parameters
        ----------
        n
            The forecast horizon
        target
            The target TimeSeries used during training or passed to prediction as `series`
        past_covariate
            Optionally, the past covariates used for training.
        future_covariate
            Optionally, the future covariates used for training.
        encode_past
            Whether or not to apply encoders for past covariates
        encode_future
            Whether or not to apply encoders for future covariates

        Returns
        -------
        Tuple[past_covariate, future_covariate]
            The past_covariate and/or future_covariate for prediction/inference including the encodings.
            If input {x}_covariate is None and no {x}_encoders are given, will return `None`
            for the {x}_covariate.
        """

        return self._launch_encoder(target=target,
                                    past_covariate=past_covariate,
                                    future_covariate=future_covariate,
                                    n=n,
                                    encode_past=encode_past,
                                    encode_future=encode_future)

    def _launch_encoder(self,
                        target: Sequence[TimeSeries],
                        past_covariate: SupportedTimeSeries,
                        future_covariate: SupportedTimeSeries,
                        n: Optional[int] = None,
                        encode_past: bool = True,
                        encode_future: bool = True) -> Tuple[Sequence[TimeSeries], Sequence[TimeSeries]]:
        """Launches the encode sequence for past covariate and future covariate for either training or
        inference/prediction.

        If `n` is `None` it is a prediction, otherwise it is training.
        """

        if not self.encoding_available:
            return past_covariate, future_covariate

        target = [target] if isinstance(target, TimeSeries) else target

        if self.past_encoders and encode_past:
            past_covariate = self._encode_sequence(encoders=self.past_encoders,
                                                   transformer=self.past_transformer,
                                                   target=target,
                                                   covariate=past_covariate,
                                                   n=n)

        if self.future_encoders and encode_future:
            future_covariate = self._encode_sequence(encoders=self.future_encoders,
                                                     transformer=self.future_transformer,
                                                     target=target,
                                                     covariate=future_covariate,
                                                     n=n)
        return past_covariate, future_covariate

    def _encode_sequence(self,
                         encoders: Sequence[SingleEncoder],
                         transformer: Optional[SequentialEncoderTransformer],
                         target: Sequence[TimeSeries],
                         covariate: Optional[SupportedTimeSeries],
                         n: Optional[int] = None) -> List[TimeSeries]:
        """Sequentially encodes the index of all input target/covariate TimeSeries

        If `n` is `None` it is a prediction and method `encoder.encode_inference()` is called.
        Otherwise, it is a training case and `encoder.encode_train()` is called.
        """

        encode_method = 'encode_train' if n is None else 'encode_inference'

        encoded_sequence = []
        if covariate is None:
            covariate = [None] * len(target)
        else:
            covariate = [covariate] if isinstance(covariate, TimeSeries) else covariate

        for ts, pc in zip(target, covariate):
            encoded = concatenate([getattr(enc, encode_method)(target=ts,
                                                               covariate=pc,
                                                               merge_covariate=False,
                                                               n=n) for enc in encoders], axis=DIMS[1])
            encoded_sequence.append(self._merge_covariate(encoded=encoded, covariate=pc))

        if transformer is not None:
            encoded_sequence = transformer.transform(encoded_sequence)

        return encoded_sequence

    @property
    def future_encoders(self) -> List[SingleEncoder]:
        """Returns the future covariate encoder objects"""
        return self._future_encoders

    @property
    def past_encoders(self) -> List[SingleEncoder]:
        """Returns the past covariate encoder objects"""
        return self._past_encoders

    @property
    def past_transformer(self) -> SequentialEncoderTransformer:
        """Returns the past transformer object"""
        return self._past_transformer

    @property
    def future_transformer(self) -> SequentialEncoderTransformer:
        """Returns the future transformer object"""
        return self._future_transformer

    @property
    def encoder_map(self) -> Dict:
        """Mapping between encoder identifier string (from parameters at model creations) and the corresponding
        future or past covariate encoder"""
        mapper = {
            'cyclic_past': PastCyclicEncoder,
            'cyclic_future': FutureCyclicEncoder,
            'datetime_attribute_past': PastDatetimeAttributeEncoder,
            'datetime_attribute_future': FutureDatetimeAttributeEncoder,
            'position_past': PastIntegerIndexEncoder,
            'position_future': FutureIntegerIndexEncoder,
            'custom_past': PastCallableIndexEncoder,
            'custom_future': FutureCallableIndexEncoder
        }
        return mapper

    def _setup_encoders(self, params: Dict) -> None:
        """Sets up/Initializes all past and future encoders and an optional transformer from `add_encoder` parameter
        used at model creation.


        Parameters
        ----------
        params
            Dict from parameter `add_encoders` (kwargs) used at model creation. Relevant parameters are:
            * params={'cyclic': {'past': ['month', 'dayofweek', ...], 'future': [same as for 'past']}}
        """
        past_encoders, future_encoders = self._process_input_encoders(params)

        if not past_encoders and not future_encoders:
            return

        self._past_encoders = [self.encoder_map[enc_id](self.input_chunk_length,
                                                        self.output_chunk_length,
                                                        attr) for enc_id, attr in past_encoders]
        self._future_encoders = [self.encoder_map[enc_id](self.input_chunk_length,
                                                          self.output_chunk_length,
                                                          attr) for enc_id, attr in future_encoders]
        self.encoding_available = True

    def _setup_transformer(self, params: Dict) -> None:
        """Sets up/Initializes an optional transformer from `add_encoder` parameter used at model creation.

        Parameters
        ----------
        params
            Dict from parameter `add_encoders` (kwargs) used at model creation. Relevant parameters are:
            * params={..., 'transformer': Scaler()}
        """
        transformer, transform_past_mask, transform_future_mask = self._process_input_transformer(params)
        if transform_past_mask:
            self._past_transformer = SequentialEncoderTransformer(copy.deepcopy(transformer), transform_past_mask)
        if transform_future_mask:
            self._future_transformer = SequentialEncoderTransformer(copy.deepcopy(transformer), transform_future_mask)

    def _process_input_encoders(self, params: Dict) -> Tuple[List, List]:
        """Processes input and returns two lists of tuples `(encoder_id, attribute)` from relevant encoder
        parameters at model creation.

        Parameters
        ----------
        params
            The `add_encoders` dict used at model creation. Must follow this convention:
                `{encoder keyword: {temporal keyword: List[attributes]}}`

            Tuples of `(encoder_id, attribute)` are extracted from `add_encoders` to instantiate the `SingleEncoder`
            objects:
            * The `encoder_id` is extracted as follows:
                str(encoder_kw) + str(temporal_kw) -> 'cyclic' + 'past' -> `encoder_id` = 'cyclic_past'
                The `encoder_id` is used to map the parameters with the corresponding `SingleEncoder` objects.
            * The `attribute` is extracted from the values given by values under `temporal_kw`
                `attribute` = 'month'
                ...
                The `attribute` tells the `SingleEncoder` which attribute of the index to encode

        Raises
        ------
        ValueError
            1) if the outermost key is other than (`past`, `future`, `absolute`)
            2) if the innermost values are other than type `str` or `Sequence`
        """

        if not params:
            return [], []

        # check input for invalid encoder types
        invalid_encoders = [enc for enc in params if not enc in ENCODER_KEYS + TRANSFORMER_KEYS]
        raise_if(len(invalid_encoders) > 0,
                 f'Encountered invalid encoder types `{invalid_encoders}` in `add_encoders` parameter at model '
                 f'creation. Supported encoder types are: `{ENCODER_KEYS + TRANSFORMER_KEYS}`.',
                 logger)

        encoders = {enc: params.get(enc, None) for enc in ENCODER_KEYS if params.get(enc, None)}

        # check input for invalid temporal types
        invalid_time_params = list()
        for encoder, t_types in encoders.items():
            invalid_time_params += [t_type for t_type in t_types.keys() if t_type not in VALID_TIME_PARAMS]

        raise_if(len(invalid_time_params) > 0,
                 f'Encountered invalid temporal types `{invalid_time_params}` in `add_encoders` parameter at model '
                 f'creation. Supported temporal types are: `{VALID_TIME_PARAMS}`.',
                 logger)

        # convert into tuples of (encoder string identifier, encoder attribute)
        past_encoders, future_encoders = list(), list()
        for enc, enc_params in encoders.items():
            for enc_time, enc_attr in enc_params.items():
                raise_if_not(isinstance(enc_attr, VALID_ENCODER_DTYPES),
                             f'Encountered value `{enc_attr}` of invalid type `{type(enc_attr)}` for encoder '
                             f'`{enc}` in `add_encoders` at model creation. Supported data types are: '
                             f'`{VALID_ENCODER_DTYPES}`.',
                             logger)
                attrs = [enc_attr] if isinstance(enc_attr, str) else enc_attr
                for attr in attrs:
                    encoder_id = '_'.join([enc, enc_time])
                    if enc_time == PAST:
                        past_encoders.append((encoder_id, attr))
                    else:
                        future_encoders.append((encoder_id, attr))

        for temp_enc, takes_temp, temp in [(past_encoders, self.takes_past_covariates, 'past'),
                                           (future_encoders, self.takes_future_covariates, 'future')]:
            if temp_enc and not takes_temp:
                logger.warning(f'Specified {temp} encoders in `add_encoders` at model creation but model does not '
                               f'accept {temp} covariates. {temp} encoders will be ignored.')

        past_encoders = past_encoders if self.takes_past_covariates else []
        future_encoders = future_encoders if self.takes_future_covariates else []
        return past_encoders, future_encoders

    def _process_input_transformer(self, params: Dict) -> Tuple[Optional[FittableDataTransformer], List, List]:
        """Processes input params used at model creation and returns tuple of one transformer object and two masks
        that specify which past / future encoders accept being transformed.

        Parameters
        ----------
        params
            Dict from parameter `add_encoders` (kwargs) used at model creation. Relevant parameters are:
            * params={'transformer': Scaler()}
        """

        if not params:
            return None, [], []

        transformer = params.get(TRANSFORMER_KEYS[0], None)
        if transformer is None:
            return None, [], []

        raise_if_not(isinstance(transformer, VALID_TRANSFORMER_DTYPES),
                     f'Encountered `{TRANSFORMER_KEYS[0]}` of invalid type `{type(transformer)}` '
                     f'in `add_encoders` at model creation. Transformer must be an instance of '
                     f'`{VALID_TRANSFORMER_DTYPES}`.',
                     logger)

        transform_past_mask = [transform for enc in self.past_encoders for transform in enc.accept_transformer]
        transform_future_mask = [transform for enc in self.future_encoders for transform in enc.accept_transformer]
        return transformer, transform_past_mask, transform_future_mask



import unittest
import logging

from darts import TimeSeries
from darts.utils.timeseries_generation import constant_timeseries
from darts.dataprocessing import Pipeline
from darts.dataprocessing.transformers import (
    BaseDataTransformer,
    FittableDataTransformer,
    InvertibleDataTransformer,
)
from darts.dataprocessing.transformers import InvertibleMapper, Mapper


class PipelineTestCase(unittest.TestCase):
    __test__ = True

    @classmethod
    def setUpClass(cls):
        logging.disable(logging.CRITICAL)

    class DataTransformerMock1(BaseDataTransformer):
        def __init__(self):
            super().__init__()
            self.transform_called = False
            self.inverse_transform_called = False
            self.fit_called = False

        @staticmethod
        def ts_transform(data: TimeSeries) -> TimeSeries:
            return data.append_values(constant_timeseries(value=1, length=3).values())

        def transform(self, data, *args, **kwargs) -> TimeSeries:
            self.transform_called = True
            return super().transform(data, *args, **kwargs)

    class DataTransformerMock2(FittableDataTransformer, InvertibleDataTransformer):
        def __init__(self):
            super().__init__()

            self.transform_called = False
            self.inverse_transform_called = False
            self.fit_called = False

        @staticmethod
        def ts_fit(series: TimeSeries):
            pass

        @staticmethod
        def ts_transform(series: TimeSeries) -> TimeSeries:
            return series.append_values(constant_timeseries(value=2, length=3).values())

        @staticmethod
        def ts_inverse_transform(series: TimeSeries) -> TimeSeries:
            return series

        def fit(self, data):
            super().fit(data)
            self.fit_called = True
            return self

        def transform(self, data, *args, **kwargs):
            self.transform_called = True
            self.args = args
            self.kwargs = kwargs
            return super().transform(data, *args, **kwargs)

        def inverse_transform(self, data, *args, **kwargs) -> TimeSeries:
            self.inverse_transform_called = True
            self.args = args
            self.kwargs = kwargs
            return super().inverse_transform(data, *args, **kwargs)

    class PlusTenTransformer(InvertibleDataTransformer):
        def __init__(self, name="+10 transformer"):
            super().__init__(name=name)

        @staticmethod
        def ts_transform(series: TimeSeries) -> TimeSeries:
            return series.map(lambda x: x + 10)

        @staticmethod
        def ts_inverse_transform(series: TimeSeries) -> TimeSeries:
            return series.map(lambda x: x - 10)

    class TimesTwoTransformer(InvertibleDataTransformer):
        def __init__(self):
            super().__init__(name="*2 transformer")

        @staticmethod
        def ts_transform(data: TimeSeries) -> TimeSeries:
            return data.map(lambda x: x * 2)

        @staticmethod
        def ts_inverse_transform(data: TimeSeries) -> TimeSeries:
            return data.map(lambda x: x / 2)

    def test_transform(self):
        # given
        mock1 = self.DataTransformerMock1()
        mock2 = self.DataTransformerMock2()
        data = constant_timeseries(value=0, length=3)
        transformers = [mock1] * 10 + [mock2] * 10
        p = Pipeline(transformers)
        # when
        p.fit(data)
        transformed = p.transform(data)

        # then
        self.assertEqual(63, len(transformed))
        self.assertEqual([0] * 3 + [1] * 30 + [2] * 30, list(transformed.values()))
        for t in transformers:
            self.assertTrue(t.transform_called)
            self.assertFalse(t.inverse_transform_called)

    def test_inverse_raise_exception(self):
        # given
        mock = self.DataTransformerMock1()
        p = Pipeline([mock])

        # when & then
        with self.assertRaises(ValueError):
            p.inverse_transform(None)

    def test_transformers_not_modified(self):
        # given
        mock = self.DataTransformerMock1()
        p = Pipeline([mock], copy=True)

        # when
        p.transform(constant_timeseries(value=1, length=10))

        # then
        self.assertFalse(mock.transform_called)

    def test_fit(self):
        # given
        data = constant_timeseries(value=0, length=3)
        transformers = [self.DataTransformerMock1() for _ in range(10)] + [
            self.DataTransformerMock2() for _ in range(10)
        ]
        p = Pipeline(transformers)

        # when
        p.fit(data)

        # then
        for i in range(10):
            self.assertFalse(transformers[i].fit_called)
        for i in range(10, 20):
            self.assertTrue(transformers[i].fit_called)

    def test_fit_skips_superfluous_transforms(self):
        # given
        data = constant_timeseries(value=0, length=100)
        transformers = (
            [self.DataTransformerMock1() for _ in range(10)]
            + [self.DataTransformerMock2()]
            + [self.DataTransformerMock1() for _ in range(10)]
        )
        p = Pipeline(transformers)

        # when
        p.fit(data)

        # then
        for i in range(10):
            self.assertTrue(transformers[i].transform_called)
        self.assertTrue(transformers[10].fit_called)
        self.assertFalse(transformers[10].transform_called)
        for i in range(11, 21):
            self.assertFalse(transformers[i].transform_called)

    def test_transform_fit(self):
        # given
        data = constant_timeseries(value=0, length=3)
        transformers = [self.DataTransformerMock1() for _ in range(10)] + [
            self.DataTransformerMock2() for _ in range(10)
        ]
        p = Pipeline(transformers)

        # when
        _ = p.fit_transform(data)

        # then
        for t in transformers:
            self.assertTrue(t.transform_called)
        for i in range(10):
            self.assertFalse(transformers[i].fit_called)
        for i in range(10, 20):
            self.assertTrue(transformers[i].fit_called)

    def test_inverse_transform(self):
        # given
        data = constant_timeseries(value=0.0, length=3)

        transformers = [self.PlusTenTransformer(), self.TimesTwoTransformer()]
        p = Pipeline(transformers)

        # when
        transformed = p.transform(data)
        back = p.inverse_transform(transformed)

        # then
        self.assertEqual(data, back)

    def test_getitem(self):
        # given

        transformers = [
            self.PlusTenTransformer(name="+10 transformer{}".format(i))
            for i in range(0, 10)
        ]
        p = Pipeline(transformers)

        # when & then
        # note : only compares string representations, since __getitem__() copies the transformers
        self.assertEqual(str(p[1]._transformers), str([transformers[1]]))
        self.assertEqual(str(p[4:8]._transformers), str(transformers[4:8]))

        with self.assertRaises(ValueError):
            p["invalid attempt"]

    def test_raises_on_non_transformers(self):
        # given
        input_list = list(range(10))

        # when & then
        with self.assertRaises(
            ValueError,
            msg="transformers should be objects deriving from BaseDataTransformer",
        ):
            Pipeline(input_list)

    def test_raises_on_bad_key(self):
        # given
        bad_key = 12.0
        p = Pipeline([])

        # when & then
        with self.assertRaises(ValueError, msg="Key must be int, str or slice"):
            p[bad_key]

    def test_multi_ts(self):

        series1 = constant_timeseries(value=0.0, length=3)
        series2 = constant_timeseries(value=1.0, length=3)

        data = [series1, series2]

        mapper1 = InvertibleMapper(fn=lambda x: x + 10, inverse_fn=lambda x: x - 10)
        mapper2 = InvertibleMapper(fn=lambda x: x * 10, inverse_fn=lambda x: x / 10)

        transformers = [mapper1, mapper2]
        p = Pipeline(transformers)

        # when
        transformed = p.transform(data)
        back = p.inverse_transform(transformed)

        # then
        self.assertEqual(data, back)

    def test_pipeline_partial_inverse(self):
        series = constant_timeseries(value=0.0, length=3)

        def plus_ten(x):
            return x + 10

        mapper = Mapper(fn=plus_ten)
        mapper_inv = InvertibleMapper(fn=lambda x: x + 2, inverse_fn=lambda x: x - 2)

        series_plus_ten = mapper.transform(series)

        pipeline = Pipeline([mapper, mapper_inv])

        transformed = pipeline.transform(series)

        # should fail, since partial is False by default
        with self.assertRaises(ValueError):
            pipeline.inverse_transform(transformed)

        back = pipeline.inverse_transform(transformed, partial=True)

        # while the +/- 2 is inverted, the +10 operation is not
        self.assertEqual(series_plus_ten, back)

    def test_pipeline_verbose(self):
        """
        Checks if the verbose param applied to the pipeline is changing the verbosity level in the
        contained transformers.
        """

        def plus_ten(x):
            return x + 10

        mapper = Mapper(fn=plus_ten, verbose=True)
        mapper_inv = InvertibleMapper(
            fn=lambda x: x + 2, inverse_fn=lambda x: x - 2, verbose=True
        )

        verbose_value = False
        pipeline = Pipeline([mapper, mapper_inv], verbose=verbose_value)

        for transformer in pipeline:
            self.assertEqual(transformer._verbose, verbose_value)

    def test_pipeline_n_jobs(self):
        """
        Checks if the n_jobs param applied to the pipeline is changing the verbosity level in the
        contained transformers.
        """

        def plus_ten(x):
            return x + 10

        mapper = Mapper(fn=plus_ten, n_jobs=1)
        mapper_inv = InvertibleMapper(
            fn=lambda x: x + 2, inverse_fn=lambda x: x - 2, n_jobs=2
        )

        n_jobs_value = -1
        pipeline = Pipeline([mapper, mapper_inv], n_jobs=n_jobs_value)

        for transformer in pipeline:
            self.assertEqual(transformer._n_jobs, n_jobs_value)



"""
Missing Values Filler
---------------------
"""
from typing import Union, Sequence, List
from darts import TimeSeries
from darts.dataprocessing.transformers import BaseDataTransformer
from darts.utils.missing_values import fill_missing_values
from darts.logging import get_logger, raise_if, raise_if_not

logger = get_logger(__name__)


class MissingValuesFiller(BaseDataTransformer):
    def __init__(self,
                 fill: Union[str, float] = 'auto',
                 name: str = "MissingValuesFiller",
                 n_jobs: int = 1,
                 verbose: bool = False):
        """
        Data transformer to fill missing values from a (sequence of) deterministic TimeSeries.

        Parameters
        ----------
        fill
            The value used to replace the missing values.
            If set to 'auto', will auto-fill missing values using the `pandas.Dataframe.interpolate()` method.
        name
            A specific name for the transformer
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a `Sequence[TimeSeries]` is
            passed as input to a method, parallelising operations regarding different `TimeSeries`. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        verbose
            Optionally, whether to print operations progress
        """
        raise_if_not(isinstance(fill, str) or isinstance(fill, float),
                     "`fill` should either be a string or a float",
                     logger)
        raise_if(isinstance(fill, str) and fill != 'auto',
                 "invalid string for `fill`: can only be set to 'auto'",
                 logger)

        super().__init__(name=name, n_jobs=n_jobs, verbose=verbose)
        self._fill = fill

    @staticmethod
    def ts_transform(series: TimeSeries, fill: Union[str, float], **kwargs) -> TimeSeries:
        return fill_missing_values(series, fill, **kwargs)

    def transform(self,
                  series: Union[TimeSeries, Sequence[TimeSeries]],
                  *args, **kwargs) -> Union[TimeSeries, List[TimeSeries]]:
        # adding the fill param
        return super().transform(series, self._fill, *args, **kwargs)



"""
darts
-----
"""

from .timeseries import TimeSeries, concatenate
import matplotlib as mpl
from matplotlib import cycler


__version__ = '0.15.0'

colors = cycler(color=['black', '003DFD', 'b512b8', '11a9ba', '0d780f', 'f77f07', 'ba0f0f'])

u8plots_mplstyle = {
    'font.family': 'sans serif',
    'axes.edgecolor': 'black',
    'axes.grid': True,
    'axes.labelcolor': '#333333',
    'axes.labelweight': 600,
    'axes.linewidth': 1,
    'axes.prop_cycle': colors,
    'axes.spines.top': False,
    'axes.spines.right': False,
    'axes.spines.bottom': False,
    'axes.spines.left': False,
    'grid.color': '#dedede',
    'legend.frameon': False,
    'lines.linewidth': 1.3,
    'xtick.color': '#333333',
    'xtick.labelsize': 'small',
    'ytick.color': '#333333',
    'ytick.labelsize': 'small',
    'xtick.bottom': False,
}


mpl.rcParams.update(u8plots_mplstyle)



import numpy as np
from typing import Callable, Union
import copy

from .window import Window, CRWindow, NoWindow
from .cost_matrix import CostMatrix
from darts import TimeSeries
from darts.logging import get_logger, raise_if_not, raise_if

logger = get_logger(__name__)

SeriesValue = Union[np.ndarray, np.floating]
DistanceFunc = Callable[[SeriesValue, SeriesValue], float]


# CORE ALGORITHM
def _dtw_cost_matrix(
    x: np.ndarray, y: np.ndarray, dist: DistanceFunc, window: Window
) -> np.ndarray:

    dtw = CostMatrix._from_window(window)

    dtw.fill(np.inf)
    dtw[0, 0] = 0

    for i, j in window:
        cost = dist(x[i - 1], y[j - 1])
        min_cost_prev = min(dtw[i - 1, j], dtw[i, j - 1], dtw[i - 1, j - 1])
        dtw[i, j] = cost + min_cost_prev

    return dtw


def _dtw_path(dtw: CostMatrix) -> np.ndarray:
    i = dtw.n
    j = dtw.m

    path = []

    while i > 0 or j > 0:
        path.append((i - 1, j - 1))

        stencil = [
            (i - 1, j - 1),  # diagonal
            (i - 1, j),  # left
            (i, j - 1),  # down
        ]

        costs = [dtw[i, j] for i, j in stencil]
        index_min = costs.index(min(costs))

        i, j = stencil[index_min]

    path.reverse()
    return np.array(path)


# MULTI-SCALE
def _down_sample(high_res: np.ndarray):
    needs_padding = len(high_res) & 1
    if needs_padding:
        high_res = np.append(high_res, high_res[-1])

    low_res = np.reshape(high_res, (-1, 2))
    low_res = np.mean(low_res, axis=1)

    return low_res


def _expand_window(low_res_path: np.ndarray, n: int, m: int, radius: int) -> CRWindow:
    high_res_grid = CRWindow(n, m)

    def is_valid(cell):
        valid_x = 1 <= cell[0] <= n
        valid_y = 1 <= cell[1] <= m

        return valid_x and valid_y

    # up-sample path, 1 grid point computed at half-resolution becomes at least 4
    # (1,2) and (2,1) allows for better DWT solution, as it isn't diagonally constrained
    #
    # 4-sample         6-sample
    #     X X             X X
    #     X X    vs.    X X X
    # X X             X X X
    # X X             X X

    pattern = [(0, 0, 2), (1, 0, 3), (2, 1, 2)]

    for i, j in low_res_path:
        for column, start, end in pattern:
            # +1 offset since x[0], y[0] maps to DTW[1][1]

            column += i * 2 + 1

            # ensure no out of bounds
            # enlarge vertically by radius is O(1), due to compressed row representation
            start = max(1, min(m + 1, start + j * 2 - radius))
            end = max(1, min(m + 1, end + j * 2 + 1 + radius))

            # to create an nxn block, add (0-n) range n times
            # ensure no out of bounds
            #   if column is too large m-column -> 0
            #   if column-k is too small, column < radius+1
            for k in range(0, min(radius + 1, column, n - column + 1)):
                high_res_grid.add_range(column - k, start, end)

    return high_res_grid


def _fast_dtw(
    x: np.ndarray, y: np.ndarray, dist: DistanceFunc, radius: int, depth: int = 0
) -> CostMatrix:
    n = len(x)
    m = len(y)
    min_size = radius + 2

    if n < min_size or m < min_size or radius == -1:
        window = NoWindow()
        window.init_size(n, m)
        cost = _dtw_cost_matrix(x, y, dist, window)

        return cost

    half_x = _down_sample(x)
    half_y = _down_sample(y)

    low_res_cost = _fast_dtw(half_x, half_y, dist, radius, depth + 1)
    low_res_path = _dtw_path(low_res_cost)
    window = _expand_window(low_res_path, len(x), len(y), radius)
    cost = _dtw_cost_matrix(x, y, dist, window)

    return cost


# Public API Functions
class DTWAlignment:
    n: int
    m: int
    series1: TimeSeries
    series2: TimeSeries
    cost: CostMatrix

    def __init__(self, series1: TimeSeries, series2: TimeSeries, cost: CostMatrix):

        self.n = len(series1)
        self.m = len(series2)
        self.series1 = series1
        self.series2 = series2
        self.cost = cost

    from ._plot import plot, plot_alignment

    def path(self) -> np.ndarray:
        """
        Returns
        -------
        np.ndarray of shape `(len(path), 2)`
            An array of indices [[i0,j0], [i1,j1], [i2,j2], ...], where i indexes into series1
            and j indexes into series2.
            Indices are in monotonic order, path[n] >= path[n-1]
        """

        if hasattr(self, "_path"):
            return self._path
        self._path = _dtw_path(self.cost)
        return self._path

    def distance(self) -> float:
        """
        Returns
        -------
        float
            The total distance between pair-wise elements in the two series after warping.
        """
        return self.cost[(self.n, self.m)]

    def mean_distance(self) -> float:
        """
        Returns
        -------
        float
            The mean distance between pair-wise elements in the two series after warping.
        """
        if hasattr(self, "_mean_distance"):
            return self._mean_distance

        path = self.path()
        self._mean_distance = self.distance() / len(path)
        return self._mean_distance

    def warped(self) -> (TimeSeries, TimeSeries):
        """
        Warps the two time series according to the warp path returned by .path(), which minimizes
        the pair-wise distance.
        This will bring two time series that are out-of-phase back into phase.

        Returns
        -------
        (TimeSeries, TimeSeries)
            Two new TimeSeries instances of the same length, indexed by pd.RangeIndex.
        """

        series1 = self.series1
        series2 = self.series2

        xa1 = series1.data_array(copy=False)
        xa2 = series2.data_array(copy=False)

        path = self.path()

        warped_series1 = xa1[path[:, 0]]
        warped_series2 = xa2[path[:, 1]]

        time_dim1 = series1._time_dim
        time_dim2 = series2._time_dim

        range_index = True

        if range_index:
            warped_series1 = warped_series1.reset_index(dims_or_levels=time_dim1)
            warped_series2 = warped_series2.reset_index(dims_or_levels=time_dim2)

        # todo: prevent time information being lost after warping
        # Applying time index from series1 to series2 (take_dates = True) is disabled for consistency reasons
        # Applying the warp path to the dates directly will result in duplicate dates
        # and hence values being lost when converting back to a TimeSeries.
        # As a result series1 will not be warped, whereas series2 will be.
        # It could also cause the two series to have different lengths, if len(series1) < len(series2)
        # One could generate intermediate dates, but then the series would not have a consistent frequency

        take_dates = False
        if take_dates:
            time_index = warped_series1[time_dim1]
            time_index = time_index.rename({time_dim1: time_dim2})
            warped_series2[time_dim2] = time_index

        return TimeSeries.from_xarray(warped_series1), TimeSeries.from_xarray(
            warped_series2
        )


def default_distance_multi(x_values: np.ndarray, y_values: np.ndarray):
    return np.sum(np.abs(x_values - y_values))


def default_distance_uni(x_value: float, y_value: float):
    return abs(x_value - y_value)


def dtw(
    series1: TimeSeries,
    series2: TimeSeries,
    window: Window = NoWindow(),
    distance: Union[DistanceFunc, None] = None,
    multi_grid_radius: int = -1,
) -> DTWAlignment:
    """
    Determines the optimal alignment between two time series series1 and series2,
    according to the Dynamic Time Warping algorithm.
    The alignment minimizes the distance between pair-wise elements after warping.
    All elements in the two series are matched and are in strictly monotonically increasing order.
    Considers only the values in the series, ignoring the time axis.

    Dynamic Time Warping can be applied to determine how closely two time series correspond,
    irrespective of phase, length or speed differences.

    Parameters
    ----------
    series1
        `TimeSeries`
    series2
        A `TimeSeries`
    window
        Used to constrain the search for the optimal alignment: see SakoeChiba and Itakura.
        Default considers all possible alignments.
    distance
        Function taking as input either two `floats` for univariate series or two `np.ndarray`,
        and returning the distance between them.

        Defaults to the abs difference for univariate-data and the
        sum of the abs difference for multi-variate series.
    multi_grid_radius
        Default radius of -1 results in an exact evaluation of the dynamic time warping algorithm.
        Without constraints DTW runs in O(nxm) time where n,m are the size of the series.
        Exact evaluation with no constraints, will result in a performance warning on large datasets.

        Setting multi_grid_radius to a value other than -1, will enable the approximate multi-grid solver,
        which executes in linear time, vs quadratic time for exact evaluation.
        Increasing radius trades solution accuracy for performance.

    Returns
    -------
    DTWAlignment
        Helper object for getting warp path, mean_distance, distance and warped time series
    """

    if (
        multi_grid_radius == -1
        and type(window) is NoWindow
        and len(series1) * len(series2) > 10 ** 6
    ):
        logger.warn(
            "Exact evaluation will result in poor performance on large datasets."
            " Consider enabling multi-grid or using a window."
        )

    both_univariate = series1.is_univariate and series2.is_univariate

    if distance is None:
        raise_if_not(
            series1.n_components == series2.n_components,
            "Expected series to have same number of components, or to supply custom distance function",
            logger,
        )

        distance = default_distance_uni if both_univariate else default_distance_multi

    if both_univariate:
        values_x = series1.univariate_values(copy=False)
        values_y = series2.univariate_values(copy=False)
    else:
        values_x = series1.values(copy=False)
        values_y = series2.values(copy=False)

    raise_if(
        np.any(np.isnan(values_x)),
        "Dynamic Time Warping does not support nan values. "
        "You can use the module darts.utils.missing_values to fill them, "
        "before passing them to dtw.",
        logger,
    )
    raise_if(
        np.any(np.isnan(values_y)),
        "Dynamic Time Warping does not support nan values. "
        "You can use the module darts.utils.missing_values to fill them,"
        "before passing it into dtw",
        logger,
    )

    window = copy.deepcopy(window)
    window.init_size(len(values_x), len(values_y))

    raise_if(multi_grid_radius < -1, "Expected multi-grid radius to be positive or -1")

    if multi_grid_radius >= 0:
        raise_if_not(
            isinstance(window, NoWindow),
            "Multi-grid solver does not currently support windows",
            logger,
        )
        cost_matrix = _fast_dtw(values_x, values_y, distance, multi_grid_radius)
    else:
        cost_matrix = _dtw_cost_matrix(values_x, values_y, distance, window)

    return DTWAlignment(series1, series2, cost_matrix)



"""
Regression Model
----------------
A `RegressionModel` forecasts future values of a target series based on lagged values of

* The target series (past lags only)

* An optional past_covariates series (past lags only)

* An optional future_covariates series (possibly past and future lags)


The regression models are learned in a supervised way, and they can wrap around any "scikit-learn like" regression model
acting on tabular data having `fit()` and `predict()` functions.

Darts also provides `LinearRegressionModel` and `RandomForest`, which are regression models wrapping around
scikit-learn linear regression and random forest regression, respectively.

Behind the scenes this model is tabularizing the time series data to make it work with regression models.

The lags can be specified either using an integer - in which case it represents the _number_ of (past or future) lags
to take into consideration, or as a list - in which case the lags have to be enumerated (strictly negative values
denoting past lags and positive values including 0 denoting future lags).
"""
from typing import Union, Sequence, Optional, Tuple, List, Dict
import numpy as np

from darts.timeseries import TimeSeries
from sklearn.linear_model import LinearRegression
from sklearn.multioutput import MultiOutputRegressor
from darts.models.forecasting.forecasting_model import GlobalForecastingModel
from darts.logging import raise_if, raise_if_not, get_logger, raise_log
from darts.utils.data.sequential_dataset import MixedCovariatesSequentialDataset
from darts.utils.data.inference_dataset import MixedCovariatesInferenceDataset


logger = get_logger(__name__)


def _shift_matrices(past_matrix: Optional[np.ndarray],
                    future_matrix: np.ndarray) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Given two matrices, consumes the first column of the past_matrix (if available), and moves the first column of
    the future_matrix to the end of the past_matrix. This can be seen as a time shift, sending the new past value
    to the past_matrix. `None` will be returned in case a matrix is empty after performing the operation.
    """
    raise_if(future_matrix is None,
             "Future matrix to be shifted cannot be None")

    new_past_matrix = []

    # if more than one timestep in past_matrix add all except for the oldest timestep to new_past_matrix
    if past_matrix.shape[1] > 1:
        new_past_matrix.append(past_matrix[:, 1:, :])

    first_future = future_matrix[:, 0, :]
    first_future = first_future.reshape(first_future.shape[0], 1, first_future.shape[1])
    new_past_matrix.append(first_future)
    future_matrix = future_matrix[:, 1:, :] if future_matrix.shape[1] > 1 else None
    new_past_matrix = np.concatenate(new_past_matrix, axis=1)

    return new_past_matrix, future_matrix


def _update_min_max(current_min: Union[int, None], current_max: Union[int, None],
                    new_values: Sequence[int]) -> Tuple[int, int]:
    """
    Helper function that, given min, max and some values, updates min and max. Min and max can be set to None, if not
    set yet.
    """
    new_min = min(min(new_values), current_min) if current_min is not None else min(new_values)
    new_max = max(max(new_values), current_max) if current_max is not None else max(new_values)
    return new_min, new_max


class RegressionModel(GlobalForecastingModel):
    def __init__(self,
                 lags: Union[int, list] = None,
                 lags_past_covariates: Union[int, List[int]] = None,
                 lags_future_covariates: Union[Tuple[int, int], List[int]] = None,
                 model=None):
        """Regression Model
        Can be used to fit any scikit-learn-like regressor class to predict the target time series from lagged values.

        Parameters
        ----------
        lags
            Lagged target values used to predict the next time step. If an integer is given the last `lags` past lags
            are used (from -1 backward). Otherwise a list of integers with lags is required (each lag must be < 0).
        lags_past_covariates
            Number of lagged past_covariates values used to predict the next time step. If an integer is given the last
            `lags_past_covariates` past lags are used (inclusive, starting from lag -1). Otherwise a list of integers
            with lags < 0 is required.
        lags_future_covariates
            Number of lagged future_covariates values used to predict the next time step. If an tuple (past, future) is
            given the last `past` lags in the past are used (inclusive, starting from lag -1) along with the first
            `future` future lags (starting from 0 - the prediction time - up to `future - 1` included). Otherwise a list
            of integers with lags is required.
        model
            Scikit-learn-like model with `fit()` and `predict()` methods. Also possible to use model that doesn't
            support multi-output regression for multivariate timeseries, in which case one regressor
            will be used per component in the multivariate series.
            If None, defaults to: `sklearn.linear_model.LinearRegression(n_jobs=-1)`
        """

        super().__init__()

        self.model = model
        self.lags = None
        self.lags_past_covariates = None
        self.lags_historical_covariates = None
        self.lags_future_covariates = None
        self.min_lag = None
        self.max_lag = None

        # model checks
        if self.model is None:
            self.model = LinearRegression(n_jobs=-1)

        if not callable(getattr(self.model, "fit", None)):
            raise_log(Exception("Provided model object must have a fit() method", logger))
        if not callable(getattr(self.model, "predict", None)):
            raise_log(Exception("Provided model object must have a predict() method", logger))

        # lags checks and processing to arrays
        raise_if((lags is None) and (lags_future_covariates is None) and (lags_past_covariates is None),
                 "At least one of `lags`, `lags_future_covariates` or `lags_past_covariates` must be not None.")

        lags_type_checks = [
            (lags, 'lags'),
            (lags_past_covariates, 'lags_past_covariates'),
        ]

        # checking types
        for _lags, lags_name in lags_type_checks:
            raise_if_not(isinstance(_lags, (int, list)) or _lags is None,
                         f"`{lags_name}` must be of type int or list. Given: {type(_lags)}.")

            raise_if(isinstance(lags, bool),
                     f"`{lags_name}` must be of type int or list, not bool.")

        raise_if_not(isinstance(lags_future_covariates, (tuple, list)) or lags_future_covariates is None,
                     f"`lags_future_covariates` must be of type tuple or list. Given: {type(lags_future_covariates)}.")

        raise_if(isinstance(lags_future_covariates, bool),
                 "`lags_future_covariates` must be of type int or list, not bool.")

        if isinstance(lags_future_covariates, tuple):
            raise_if_not(
                len(lags_future_covariates) == 2 and isinstance(lags_future_covariates[0], int)
                and isinstance(lags_future_covariates[1], int),
                "`lags_future_covariates` tuple must be of length 2, and must contain two integers"
            )
            raise_if(isinstance(lags_future_covariates[0], bool) or isinstance(lags_future_covariates[1], bool),
                     "`lags_future_covariates` tuple must contain intergers, not bool")

        if isinstance(lags, int):
            raise_if_not(
                lags > 0,
                f"`lags` must be strictly positive. Given: {lags}.")
            # selecting last `lags` lags, starting from position 1 (skipping current, pos 0, the one we want to predict)
            self.lags = list(range(-lags, 0))

        elif isinstance(lags, list):
            for lag in lags:
                raise_if(not isinstance(lag, int) or (lag >= 0),
                         f"Every element of `lags` must be a strictly negative integer. Given: {lags}.")
            self.lags = sorted(lags)

        if isinstance(lags_past_covariates, int):
            raise_if_not(
                lags_past_covariates > 0,
                f"`lags_past_covariates` must be an integer > 0. Given: {lags_past_covariates}.")
            self.lags_past_covariates = list(range(-lags_past_covariates, 0))

        elif isinstance(lags_past_covariates, list):
            for lag in lags_past_covariates:
                raise_if(
                    not isinstance(lag, int) or (lag >= 0),
                    f"Every element of `lags_covariates` must be an integer < 0. Given: {lags_past_covariates}.")
            self.lags_past_covariates = sorted(lags_past_covariates)

        if isinstance(lags_future_covariates, tuple):
            raise_if_not(
                lags_future_covariates[0] >= 0 and lags_future_covariates[1] >= 0,
                f"`lags_past_covariates` tuple must contain integers >= 0. Given: {lags_future_covariates}.")

            if lags_future_covariates[0] is not None and lags_future_covariates[0] > 0:
                self.lags_historical_covariates = list(range(-lags_future_covariates[0], 0))
            if lags_future_covariates[1] is not None and lags_future_covariates[1] > 0:
                self.lags_future_covariates = list(range(0, lags_future_covariates[1]))

        elif isinstance(lags_future_covariates, list):
            for lag in lags_future_covariates:
                raise_if(
                    not isinstance(lag, int) or isinstance(lag, bool),
                    f"Every element of `lags_future_covariates` must be an integer. Given: {lags_future_covariates}.")

            lags_future_covariates = np.array(sorted(lags_future_covariates))
            self.lags_historical_covariates = list(lags_future_covariates[lags_future_covariates < 0])
            self.lags_future_covariates = list(lags_future_covariates[lags_future_covariates >= 0])

            if len(self.lags_historical_covariates) == 0:
                self.lags_historical_covariates = None

            if len(self.lags_future_covariates) == 0:
                self.lags_future_covariates = None

        # retrieving min and max lags, since they will be necessary for choosing appropriate input_chunk_size and
        # output_chunk_size later

        if lags is not None:
            # min (index 0) and max (index -1) are enough since the array is already sorted
            self.min_lag, self.max_lag = _update_min_max(self.min_lag, self.max_lag, [self.lags[0], self.lags[-1]])

        lags = [
            self.lags_past_covariates,
            self.lags_historical_covariates,
            self.lags_future_covariates
        ]
        for lag in lags:
            if lag is not None:
                self.min_lag, self.max_lag = _update_min_max(
                    self.min_lag,
                    self.max_lag,
                    [lag[0], lag[-1]]
                )

    def _get_last_prediction_time(self, series, forecast_horizon, overlap_end):
        # overrides the ForecastingModel _get_last_prediction_time, taking care of future lags if any
        extra_shift = max(0, self.max_lag)

        if overlap_end:
            last_valid_pred_time = series.time_index[-1 - extra_shift]
        else:
            last_valid_pred_time = series.time_index[-forecast_horizon - extra_shift]

        return last_valid_pred_time

    def _get_training_data(self, training_dataset: MixedCovariatesSequentialDataset):
        """
        Helper function turning a MixedCovariatesSequentialDataset into training matrices X and y, like required by
        sklearn models.
        """

        training_samples = []
        training_labels = []

        for past_target, past_covariate, historic_future_covariate, future_covariate, future_target in training_dataset:
            row = []
            if self.lags is not None:
                row.append(past_target[self.lags].reshape(1, -1))
            covariates = [
                (past_covariate, self.lags_past_covariates),
                (historic_future_covariate, self.lags_historical_covariates),
                (future_covariate, self.lags_future_covariates)
            ]
            for covariate, lags in covariates:
                if lags is not None:
                    row.append(covariate[lags].reshape(1, -1))

            training_samples.append(np.concatenate(row, axis=1))
            training_labels.append(future_target[0].reshape(1, -1))

        training_samples = np.concatenate(training_samples, axis=0)
        training_labels = np.concatenate(training_labels, axis=0)
        return training_samples, training_labels

    def _create_lagged_data(self, target_series, past_covariates, future_covariates, max_samples_per_ts):
        """
        Helper function that creates training/validation matrices (X and y as required in sklearn), given series and
        max_samples_per_ts.
        """

        # setting proper input and output chunk length. We need to find all necessary lags in the dataset. Thus, we
        # need an input chunk length that contains the oldest past lag (min_lag), and we need to find the most future
        # lag in the future_covariates as well. In case no future covariate are required, we still need an output
        # chunk length of 1 for having the prediction target value.
        training_dataset = MixedCovariatesSequentialDataset(
            target_series=target_series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            input_chunk_length=max(0, -self.min_lag),
            output_chunk_length=max(1, self.max_lag + 1),  # max lag + 1 since we need to access that index
            max_samples_per_ts=max_samples_per_ts
        )

        return self._get_training_data(training_dataset=training_dataset)

    def _fit_model(self, target_series, past_covariates, future_covariates, max_samples_per_ts, **kwargs):
        """
        Function that fit the model. Deriving classes can override this method for adding additional parameters (e.g.,
        adding validation data), keeping the sanity checks on series performed by fit().
        """
        training_samples, training_labels = self._create_lagged_data(
            target_series, past_covariates, future_covariates, max_samples_per_ts)

        # if training_labels is of shape (n_samples, 1) we flatten it to have shape (n_samples,)
        if len(training_labels.shape) == 2 and training_labels.shape[1] == 1:
            training_labels = training_labels.ravel()
        self.model.fit(training_samples, training_labels, **kwargs)

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        max_samples_per_ts: Optional[int] = None,
        **kwargs
    ) -> None:
        """
        Fits/trains the model using the provided list of features time series and the target time series.

        Parameters
        ----------
        series : TimeSeries or list of TimeSeries
            TimeSeries or Sequence[TimeSeries] object containing the target values.
        past_covariates : TimeSeries or list of TimeSeries, optional
            Optionally, a series or sequence of series specifying past-observed covariates
        future_covariates : TimeSeries or list of TimeSeries, optional
            Optionally, a series or sequence of series specifying future-known covariates
        max_samples_per_ts : int, optional
            This is an integer upper bound on the number of tuples that can be produced
            per time series. It can be used in order to have an upper bound on the total size of the dataset and
            ensure proper sampling. If `None`, it will read all of the individual time series in advance (at dataset
            creation) to know their sizes, which might be expensive on big datasets.
            If some series turn out to have a length that would allow more than `max_samples_per_ts`, only the
            most recent `max_samples_per_ts` samples will be considered.
        **kwargs : dict, optional
            Additional keyword arguments passed to the `fit` method of the model.
        """
        super().fit(series=series, past_covariates=past_covariates, future_covariates=future_covariates)

        raise_if(
            past_covariates is not None and self.lags_past_covariates is None,
            "`past_covariates` not None in `fit()` method call, but `lags_past_covariates` is None in constructor")

        raise_if(
            past_covariates is None and self.lags_past_covariates is not None,
            "`past_covariates` is None in `fit()` method call, but `lags_past_covariates` is not None in"
            "constructor.")

        raise_if(
            future_covariates is not None and (self.lags_historical_covariates is None
                                               and self.lags_future_covariates is None),
            "`future_covariates` not None in `fit()` method call, but `lags_future_covariates` is None in "
            "constructor.")

        raise_if(
            future_covariates is None and (self.lags_future_covariates is not None
                                           or self.lags_historical_covariates is not None),
            "`future_covariates` is None in `fit()` method call, but `lags_future_covariates` is not None in"
            "constructor.")

        # saving the input dim, so that we can perform the dim check at prediction time
        series_dim = series.width if isinstance(series, TimeSeries) else series[0].width
        covariates_dim = 0
        for covariates in [past_covariates, future_covariates]:
            if covariates is not None:
                if isinstance(covariates, TimeSeries):
                    covariates_dim += covariates.width
                else:
                    covariates_dim += covariates[0].width

        self.input_dim = series_dim + covariates_dim

        # if series is multivariate wrap model with MultiOutputRegressor
        if not series[0].is_univariate:
            self.model = MultiOutputRegressor(self.model, n_jobs=1)

        self._fit_model(series, past_covariates, future_covariates, max_samples_per_ts, **kwargs)

    def _get_prediction_data(
        self,
        prediction_dataset: MixedCovariatesInferenceDataset
    ):
        """
        Helper function which turns a MixedCovariatesInferenceDataset into 5 matrices. The matrices simply contain all
        the samples stacked in the first dimension.
        """
        target_matrix = []
        past_covariates_matrix = []
        historic_future_covariates_matrix = []
        future_covariates_matrix = []
        future_past_covariates_matrix = []

        for (past_target, past_covariates, historic_future_covariates, future_covariates,
             future_past_covariates, _) in prediction_dataset:

            target_matrix.append(past_target)

            if past_covariates is not None:
                past_covariates_matrix.append(past_covariates)
            if future_covariates is not None:
                future_covariates_matrix.append(future_covariates)
            if historic_future_covariates is not None:
                historic_future_covariates_matrix.append(historic_future_covariates)
            if future_past_covariates is not None:
                future_past_covariates_matrix.append(future_past_covariates)

        covariates_matrices = [
            past_covariates_matrix,
            historic_future_covariates_matrix,
            future_covariates_matrix,
            future_past_covariates_matrix
        ]

        for i in range(len(covariates_matrices)):
            covariates_matrices[i] = None if len(covariates_matrices[i]) == 0 else np.asarray(covariates_matrices[i])

        return (
            np.asarray(target_matrix),
            *covariates_matrices
        )

    def predict(
        self,
        n: int,
        series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        num_samples: int = 1,
        **kwargs
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Forecasts values for `n` time steps after the end of the series.
        Parameters
        ----------
        n : int
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        series : TimeSeries or list of TimeSeries, optional
            Optionally, one or several input `TimeSeries`, representing the history of the target series whose future
            is to be predicted. If specified, the method returns the forecasts of these series. Otherwise, the method
            returns the forecast of the (single) training series.
        past_covariates : TimeSeries or list of TimeSeries, optional
            Optionally, the past-observed covariates series needed as inputs for the model.
            They must match the covariates used for training in terms of dimension and type.
        future_covariates : TimeSeries or list of TimeSeries, optional
            Optionally, the future-known covariates series needed as inputs for the model.
            They must match the covariates used for training in terms of dimension and type.
        num_samples : int, default: 1
            Currently this parameter is ignored for regression models.
        **kwargs : dict, optional
            Additional keyword arguments passed to the `predict` method of the model. Only works with
            univariate target series.
        """
        super().predict(n, series, past_covariates, future_covariates, num_samples)
        if series is None:
            # then there must be a single TS, and that was saved in super().fit as self.training_series
            raise_if(
                self.training_series is None,
                "Input series has to be provided after fitting on multiple series.",
            )
            series = self.training_series

        if past_covariates is None and self.past_covariate_series is not None:
            past_covariates = self.past_covariate_series
        if future_covariates is None and self.future_covariate_series is not None:
            future_covariates = self.future_covariate_series

        called_with_single_series = False

        if isinstance(series, TimeSeries):
            called_with_single_series = True
            series = [series]
            past_covariates = [past_covariates] if past_covariates is not None else None
            future_covariates = [future_covariates] if future_covariates is not None else None

        # check that the input sizes match
        series_dim = series[0].width
        covariates_dim = 0
        for covariates in [past_covariates, future_covariates]:
            if covariates is not None:
                covariates_dim += covariates[0].width
        in_dim = series_dim + covariates_dim

        raise_if_not(
            in_dim == self.input_dim,
            "The dimensionality of the series provided for prediction does not match the dimensionality "
            "of the series this model has been trained on. Provided input dim = {}, "
            "model input dim = {}".format(in_dim, self.input_dim),
        )

        # checking if there are enough future values in the future_covariates. MixedCovariatesInferenceDataset would
        # detect the problem in any case, but the error message would be meaningless.

        if future_covariates is not None:
            for sample in range(len(series)):
                last_req_ts = series[sample].end_time() + (max(n + self.max_lag, n)) * series[sample].freq
                raise_if_not(
                    future_covariates[sample].end_time() >= last_req_ts,
                    "When forecasting future values for a horizon n and lags_future_covariates >= 0, future_covariates"
                    "are requires to be at least `n + max_lags`, with `max_lag` being the furthest lag in the future"
                    f"For the {sample}-th sample, last future covariate timestamp is"
                    f"{future_covariates[sample].end_time()}, whereas it should be at least {last_req_ts}."
                )

        inference_dataset = MixedCovariatesInferenceDataset(
            target_series=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            n=max(n + self.max_lag, n),  # required for retrieving enough future covariates
            input_chunk_length=max(0, -self.min_lag),
            output_chunk_length=1
        )

        # all matrices have dimensions (n_series, time, n_components)
        (
            target_matrix,
            past_covariates_matrix,
            historic_future_covariates_matrix,
            future_covariates_matrix,
            future_past_covariates_matrix
        ) = self._get_prediction_data(inference_dataset)

        predictions = []

        """
        The columns of the prediction matrix need to have the same column order as during the training step, which is
        as follows: lags | lag_cov_0 | lag_cov_1 | .. where each lag_cov_X is a shortcut for
        lag_cov_X_dim_0 | lag_cov_X_dim_1 | .., that means, the lag X value of all the dimension of the covariate
        series (when multivariate).
        """
        for i in range(n):
            # building prediction matrix
            X = []
            if self.lags is not None:
                target_series = target_matrix[:, self.lags]

                X.append(target_series.reshape(len(series), -1))

            covariates_matrices = [
                (past_covariates_matrix, self.lags_past_covariates),
                (historic_future_covariates_matrix, self.lags_historical_covariates),
                (future_covariates_matrix, self.lags_future_covariates)
            ]

            for covariate_matrix, lags in covariates_matrices:
                if lags is not None:
                    covariates = covariate_matrix[:, lags]
                    X.append(covariates.reshape(len(series), -1))

            X = np.concatenate(X, axis=1)
            prediction = self.model.predict(X, **kwargs)
            # reshape to (n_series, time (always 1), n_components)
            prediction = prediction.reshape(len(series), 1, -1)
            # appending prediction to final predictions
            predictions.append(prediction)

            if i < n - 1:
                # discard oldest target
                if target_matrix is not None:
                    target_matrix = target_matrix[:, 1:] if target_matrix.shape[1] > 1 else None
                    # adding new prediction to the target series
                    if target_matrix is None:
                        target_matrix = np.asarray(prediction)
                    else:
                        target_matrix = np.concatenate([target_matrix, prediction], axis=1)

                # discarding oldest covariate
                if past_covariates_matrix is not None:
                    past_covariates_matrix, future_past_covariates_matrix = _shift_matrices(
                        past_covariates_matrix, future_past_covariates_matrix)
                if historic_future_covariates_matrix is not None:
                    historic_future_covariates_matrix, future_covariates_matrix = _shift_matrices(
                        historic_future_covariates_matrix, future_covariates_matrix)

        predictions = np.concatenate(predictions, axis=1)
        predictions = [self._build_forecast_series(row, input_tgt) for row, input_tgt in zip(predictions, series)]

        return predictions[0] if called_with_single_series else predictions

    def __str__(self):
        return self.model.__str__()



import os

from darts import TimeSeries
from darts.datasets import (
    AirPassengersDataset, AusBeerDataset, EnergyDataset,
    HeartRateDataset, IceCreamHeaterDataset, MonthlyMilkDataset,
    SunspotsDataset, TaylorDataset, TemperatureDataset,
    USGasolineDataset, WineDataset, WoolyDataset,
    GasRateCO2Dataset, MonthlyMilkIncompleteDataset
)
from darts.datasets.dataset_loaders import (
    DatasetLoadingException,
    DatasetLoaderCSV,
    DatasetLoaderMetadata,
    DatasetLoader
)
from darts.tests.base_test_class import DartsBaseTestClass

datasets = [
    AirPassengersDataset, AusBeerDataset, EnergyDataset,
    HeartRateDataset, IceCreamHeaterDataset, MonthlyMilkDataset,
    SunspotsDataset, TaylorDataset, TemperatureDataset,
    USGasolineDataset, WineDataset, WoolyDataset,
    GasRateCO2Dataset, MonthlyMilkIncompleteDataset
]

width_datasets = [
    1, 1, 28,
    1, 2, 1,
    1, 1, 1,
    1, 1, 1,
    2, 1
]

wrong_hash_dataset = DatasetLoaderCSV(
    metadata=DatasetLoaderMetadata(
        "wrong_hash",
        uri="https://raw.githubusercontent.com/unit8co/darts/master/datasets/data/AirPassengers.csv",
        hash="will fail",
        header_time="Month",
        format_time="%Y-%m"
    )
)

wrong_url_dataset = DatasetLoaderCSV(
    metadata=DatasetLoaderMetadata(
        "wrong_url",
        uri="https://AirPassengers.csv",
        hash="167ffa96204a2b47339c21eea25baf32",
        header_time="Month",
        format_time="%Y-%m"
    )
)


class DatasetLoaderTestCase(DartsBaseTestClass):
    def tearDown(self):
        # we need to remove the cached datasets between each test
        default_directory = DatasetLoader._DEFAULT_DIRECTORY
        for f in os.listdir(default_directory):
            os.remove(os.path.join(default_directory, f))
        os.rmdir(DatasetLoader._DEFAULT_DIRECTORY)

    def test_ok_dataset(self):
        for width, dataset_cls in zip(width_datasets, datasets):
            dataset = dataset_cls()
            ts: TimeSeries = dataset.load()
            self.assertEqual(ts.width, width)

    def test_hash(self):
        with self.assertRaises(DatasetLoadingException):
            wrong_hash_dataset.load()

    def test_uri(self):
        with self.assertRaises(DatasetLoadingException):
            wrong_url_dataset.load()
